import numpy as np


def get_gene_values(liger_object, gene, methylation_indices=None, log2scale=False,
                    scale_factor=10000):
    """"""

    if methylation_indices is None:
        methylation_indices = []

    gene_vals_total = []
    for idx, adata in enumerate(liger_object.adata_list):
        if adata.isbacked:
            gene_vals = _get_gene_values_disk(adata, gene)  # To be finished
        else:
            gene_vals = _get_gene_values_memory(adata, gene)

        if log2scale and idx not in methylation_indices:
            gene_vals = np.log2(gene_vals * scale_factor + 1)

        gene_vals_total.append(gene_vals)

    return np.concatenate(gene_vals_total)


def _get_gene_values_memory(adata, gene):

    # if gene is highly variable gene
    if gene in adata.var.index:
        gene_vals = np.ravel(adata[:, gene].layers['norm_data'].toarray())

    # recover gene value from raw backup
    elif gene in adata.raw.var.index:
        from sklearn.preprocessing import normalize as sp_normalize
        normalized_data = sp_normalize(adata.raw.X)
        idx = adata.raw.var.index.get_loc(gene)
        gene_vals = np.ravel(normalized_data[:, idx].toarray())

    # use 0 for not existing gene
    else:
        gene_vals = np.zeros(adata.shape[0], dtype=np.int32)
    return gene_vals


def _get_gene_values_disk():
    return None


import numpy as np
from plotnine import *


def plot_by_dataset_and_cluster(liger_object,
                                clusters=None,
                                title=None,
                                pt_size=0.3,
                                text_size=10,
                                do_shuffle=True,
                                rand_seed=1,
                                axis_labels=None,
                                do_legend=True,
                                legend_size=7,
                                return_plots=False,
                                legend_text_size=12):
    """Plot t-SNE coordinates of cells across datasets

    Generates two plots of all cells across datasets, one colored by dataset and one colored by
    cluster. These are useful for visually examining the alignment and cluster distributions,
    respectively. If clusters have not been set yet (quantileAlignSNF not called), will plot by
    single color for second plot. It is also possible to pass in another clustering (as long as
    names match those of cells).

    Parameters
    ----------
    liger_object : TYPE
        DESCRIPTION.
    clusters : TYPE, optional
        DESCRIPTION. The default is None.
    title : TYPE, optional
        DESCRIPTION. The default is None.
    pt_size : TYPE, optional
        DESCRIPTION. The default is 0.3.
    text_size : TYPE, optional
        DESCRIPTION. The default is 3.
    do_shuffle : TYPE, optional
        DESCRIPTION. The default is True.
    rand_seed : TYPE, optional
        DESCRIPTION. The default is 1.
    axis_labels : TYPE, optional
        DESCRIPTION. The default is None.
    do_legend : TYPE, optional
        DESCRIPTION. The default is True.
    legend_size : TYPE, optional
        DESCRIPTION. The default is 5.
    return_plots : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    None.

    """
    # tsne_coords = [adata.obs['tsne_coords'] for adata in liger_object.adata_list]
    tsne_df = liger_object.tsne_coords
    tsne_df['Cluster'] = np.asarray(
        np.concatenate([adata.obs['cluster'].to_numpy() for adata in liger_object.adata_list]))
    tsne_df['Cluster'] = tsne_df['Cluster'].astype('category')
    tsne_df['Dataset'] = np.concatenate(
        [np.repeat(adata.uns['sample_name'], adata.shape[0]) for adata in liger_object.adata_list])

    if do_shuffle:
        tsne_df = tsne_df.sample(frac=1, random_state=rand_seed)

    p1 = (ggplot(data=tsne_df, mapping=aes(x='tsne1', y='tsne2', color='Dataset')) +
          geom_point(size=pt_size) +
          guides(color=guide_legend(override_aes={'size': legend_size})) +
          scale_color_hue(h=15 / 360., l=0.65, s=1., color_space='husl'))

    centers = tsne_df.groupby('Cluster').agg(tsne1=('tsne1', 'median'),
                                             tsne2=('tsne2', 'median')).reset_index()

    p2 = (ggplot(data=tsne_df, mapping=aes(x='tsne1', y='tsne2', color='Cluster')) +
          geom_point(size=pt_size) +
          geom_text(data=centers, mapping=aes(label='Cluster'), color='black', size=text_size) +
          guides(color=guide_legend(override_aes={'size': legend_size})) +
          scale_color_hue(h=15 / 360., l=0.65, s=1., color_space='husl'))

    if title:
        p1 = p1 + ggtitle(title[0])
        p2 = p2 + ggtitle(title[1])

    if axis_labels:
        p1 = p1 + xlab(axis_labels[0]) + ylab(axis_labels[1])
        p2 = p2 + xlab(axis_labels[0]) + ylab(axis_labels[1])

    p1 = p1 + theme_classic(legend_text_size)
    p2 = p2 + theme_classic(legend_text_size)

    if not do_legend:
        p1 = p1 + theme(legend_position='none')
        p2 = p2 + theme(legend_position='none')

    if return_plots:
        return ([p1, p2])
    else:
        return None


from ._cal_feature import make_feature_matrix
from ._initialization import create_liger
from ._qc import calculate_mt_pct, filtering
from ._normalization import normalize
from ._select_genes import select_genes
from ._scale import scale_not_center




import h5sparse
import warnings
import numpy as np
import numexpr as ne
import matplotlib.pyplot as plt
from scipy.stats import norm
from scipy.optimize import minimize
from sklearn.preprocessing import normalize as sp_normalize
from sklearn.utils.sparsefuncs import mean_variance_axis

from .._utilities import _h5_idx_generator


def select_genes(liger_object,
                 var_thresh: float = 0.1,
                 alpha_thresh: float =0.99,
                 num_genes=None,
                 tol: float =0.0001,
                 datasets_use=None,
                 combine='union',
                 keep_unique=False,
                 capitalize=False,
                 do_plot=False,
                 cex_use=0.3,
                 chunk_size=1000):
    """Select a subset of informative genes

    This function identifies highly variable genes from each dataset and combines these gene sets
    (either by union or intersection) for use in downstream analysis. Assuming that gene
    expression approximately follows a Poisson distribution, this function identifies genes with
    gene expression variance above a given variance threshold (relative to mean gene expression).
    It also provides a log plot of gene variance vs gene expression (with a line indicating expected
    expression across genes and cells). Selected genes are plotted in green.

    Parameters
    ----------
    liger_object : liger object
        Should have already called normalize.
    var_thresh : float, optional
        Variance threshold. Main threshold used to identify variable genes. Genes with
        expression variance greater than threshold (relative to mean) are selected.
        (higher threshold -> fewer selected genes). Accepts single value or vector with separate
        var_thresh for each dataset (the default is 0.1).
    alpha_thresh : float, optional
        Alpha threshold. Controls upper bound for expected mean gene expression
        (lower threshold -> higher upper bound) (the default is 0.99).
    num_genes : int, optional
        Number of genes to find for each dataset. Optimises the value of var_thresh
        for each dataset to get this number of genes. Accepts single value or vector with same length
        as number of datasets (the default is None).
    tol : float, optional
        Tolerance to use for optimization if num.genes values passed in (the default is 0.0001).
    datasets_use : list, optional
        List of datasets to include for discovery of highly variable genes
        (the default is using all datatsets).
    combine : str, optional, 'union' or 'intersect'
        How to combine variable genes across experiments (the default is 'union').
    keep_unique : bool, optional
        Keep genes that occur (i.e., there is a corresponding column in raw_data) only
        in one dataset (the default is False).
    capitalize : bool, optional
        Capitalize gene names to match homologous genes (ie. across species)
        (the default is False).
    do_plot : bool, optional
        Display log plot of gene variance vs. gene expression for each dataset.
        Selected genes are plotted in green (the default is False).
    cex_use : float, optional
        Point size for plot (the default is 0.3).

    Returns
    -------
    liger_object : liger object
        Object with var_genes attribute.

    Examples
    --------
    >>> adata1 = AnnData(np.arange(12).reshape((4, 3)))
    >>> adata2 = AnnData(np.arange(12).reshape((4, 3)))
    >>> ligerex = pyliger.create_liger([adata1, adata2])
    >>> pyliger.normalize(ligerex)
    >>> pyliger.select_genes(ligerex) # use default selectGenes settings
    >>> pyliger.select_genes(ligerex, var_thresh=0.8) # select a smaller subset of genes
    """
    num_samples = liger_object.num_samples

    if datasets_use is None:
        datasets_use = list(range(num_samples))

    # Expand if only single var_thresh passed
    if isinstance(var_thresh, int) or isinstance(var_thresh, float):
        var_thresh = np.repeat(var_thresh, num_samples)
    if num_genes is not None:
        num_genes = np.repeat(num_genes, num_samples)

    if not np.array_equal(np.intersect1d(datasets_use, list(range(num_samples))), datasets_use):
        datasets_use = np.intersect1d(datasets_use, list(range(num_samples)))

    genes_use = np.array([])
    # Iterate through datasets
    for i in datasets_use:
        adata = liger_object.adata_list[i]
        if capitalize:
            liger_object.adata_list[i].var.index = liger_object.adata_list[i].var.index.str.upper()

        trx_per_cell = liger_object.adata_list[i].obs['nUMI'].to_numpy()
        gene_expr_mean = liger_object.adata_list[i].var['norm_mean'].to_numpy()

        # On-disk mode (set for online learning approach)
        if adata.isbacked:
            gene_expr_var = _calc_var_online(adata, gene_expr_mean, chunk_size)

        # In-memory mode
        else:
            # each gene's expression variance (across all cells)
            gene_expr_var = _calc_var_matrix(adata)

        nolan_constant = np.mean(ne.evaluate('1 / trx_per_cell'))
        alpha_thresh_corrected = alpha_thresh / adata.shape[1]
        gene_mean_upper = gene_expr_mean + norm.ppf(1 - alpha_thresh_corrected / 2) * np.sqrt(
            ne.evaluate('gene_expr_mean * nolan_constant') / adata.shape[0])
        #base_gene_lower = np.log10(ne.evaluate('gene_expr_mean * nolan_constant'))
        base_gene_lower = ne.evaluate('log10(gene_expr_mean * nolan_constant)')

        def num_var_genes(x, num_genes_des):
            # This function returns the difference between the desired number of genes and
            # the number actually obtained when thresholded on x
            y = np.sum(
                (gene_expr_var / nolan_constant) > gene_mean_upper & np.log10(gene_expr_var, dtype=np.float64) > (
                            base_gene_lower + x))
            return np.abs(num_genes_des - y)

        if num_genes is not None:
            # Optimize to find value of x which gives the desired number of genes for this dataset
            # if very small number of genes requested, var.thresh may need to exceed 1
            optimized = minimize(fun=num_var_genes, x0=[0], agrs=num_genes[i], tol=tol, bounds=[(0, 1.5)])
            var_thresh[i] = optimized.x
            if var_thresh[i].shape[0] > 1:
                warnings.warn(
                    'Returned number of genes for dataset {} differs from requested by {}. Lower tol or alpha_thresh for better results.'.format(
                        i, optimized.x.shape[0]))

        temp = var_thresh[i]
        select_gene = ne.evaluate('((gene_expr_var / nolan_constant) > gene_mean_upper) & (log10(gene_expr_var) > (base_gene_lower + temp))')
        #select_gene = ((gene_expr_var / nolan_constant) > gene_mean_upper) & (
        #            np.log10(gene_expr_var) > (base_gene_lower + var_thresh[i]))
        genes_new = adata.var.index.to_numpy()[select_gene]

        # TODO: graph needs to be improved
        if do_plot:
            plt.plot(np.log10(gene_expr_mean), np.log10(gene_expr_var))
            plt.title(liger_object.adata_list[i].uns['sample_name'])
            plt.xlabel('Gene Expression Mean (log10)')
            plt.ylabel('Gene Expression Variance (log10)')

            plt.plot(np.log10(gene_expr_mean), np.log10(gene_expr_mean) + nolan_constant, c='p')
            plt.scatter(np.log10(gene_expr_mean[select_gene]), np.log10(gene_expr_var[select_gene]), c='g')
            plt.legend('Selected genes: ' + str(len(genes_new)), loc='best')

            plt.show()

        if combine == 'union':
            genes_use = np.union1d(genes_use, genes_new)

        if combine == 'intersect':
            if genes_use.shape[0] == 0:
                genes_use = genes_new
            genes_use = np.intersect1d(genes_use, genes_new)

    if not keep_unique:
        for i in range(num_samples):
            genes_use = np.intersect1d(genes_use, liger_object.adata_list[i].var.index)

    if genes_use.shape[0] == 0:
        warnings.warn('No genes were selected; lower var_thresh values or choose "union" for combine parameter')

    liger_object.var_genes = genes_use
    for idx, adata in enumerate(liger_object.adata_list):
        var_gene_idx = liger_object.adata_list[idx].var.index.isin(genes_use).nonzero()[0]
        liger_object.adata_list[idx].uns['var_gene_idx'] = var_gene_idx
    return None


def _calc_var_online(adata, gene_expr_mean, chunk_size):
    """"""
    sum_sq_dev = np.zeros(adata.shape[1])

    file_path = './results/' + adata.uns['sample_name'] + '.hdf5'
    with h5sparse.File(file_path, 'r') as f:
        for left, right in _h5_idx_generator(chunk_size, adata.shape[0]):
            sum_sq_dev += np.sum(np.square(np.absolute(np.subtract(f['norm_data'][left:right].toarray(), gene_expr_mean))), axis=0)
    gene_expr_var = sum_sq_dev / (adata.shape[0] - 1)
    adata.var['norm_var'] = gene_expr_var

    return gene_expr_var


def _calc_var_matrix(adata):
    """"""
    #gene_expr_var = np.ravel(
    #    np.var(adata.layers['norm_data'].toarray(), axis=0, dtype=np.float64, ddof=1))

    num_sampels = adata.shape[0]
    _, gene_expr_var = mean_variance_axis(adata.layers['norm_data'], axis=0)
    gene_expr_var = ne.evaluate('gene_expr_var * num_sampels / (num_sampels-1)') # change ddof
    adata.var['norm_var'] = gene_expr_var

    return gene_expr_var



import numpy as np
import pandas as pd
import statsmodels.stats.multitest as smt
from functools import reduce
from scipy.sparse import vstack
from sklearn.preprocessing import normalize as sp_normalize
from scipy.stats import mannwhitneyu, rankdata, tiecorrect, distributions

def run_wilcoxon(liger_object,
                 compare_method,
                 data_use='all'):
    """Perform Wilcoxon rank-sum test

    Perform Wilcoxon rank-sum tests on specified dataset using given method.

    Parameters
    ----------
    liger_object : liger object
        object.
    compare_method : str, optional, 'clusters' or 'datasets'
        This indicates the metric of the test.
    data_use : str or list
        This selects which dataset(s) to use (the default is "all").

    Returns
    -------
    results : pd data frame
    """
    ### Check parameter inputs
    if compare_method not in ['datasets', 'clusters']:
        raise ValueError('Parameter *compare.method* should be either *clusters* or *datasets*.')
    if compare_method == 'datasets':
        if len(liger_object.adata_list) < 2:
            raise ValueError('Should have at least TWO inputs to compare between datasets.')
        if isinstance(data_use, list) and len(data_use) < 2:
            raise ValueError('Should have at least TWO inputs to compare between datasets.')

    ### Create feature x sample matrix
    if data_use == 'all':
        num_samples = len(liger_object.adata_list)
        sample_names = [adata.uns['sample_name'] for adata in liger_object.adata_list]
        sample_idx = list(range(num_samples))
        print('Performing Wilcoxon test on ALL datasets: {}'.format(', '.join(sample_names)))
    else:
        num_samples = len(data_use)
        sample_names = data_use
        sample_names_all = [adata.uns['sample_name'] for adata in liger_object.adata_list]
        sample_idx = [sample_names_all.index(name) for name in sample_names]
        print('Performing Wilcoxon test on GIVEN datasets: {}'.format(', '.join(sample_names)))

    # get all shared genes of every datasets
    #genes_all = [liger_object.adata_list[idx].var.index.to_numpy() for idx in sample_idx]
    genes_all = [liger_object.adata_list[idx].raw.var.index.to_numpy() for idx in sample_idx]
    genes_shared = reduce(np.intersect1d, genes_all)

    # get feature matrix, shared genes as columns and all barcodes as rows, as well as labels of clusters and datasets
    feature_matrix = []
    clusters = np.array([])
    cell_source = np.array([])
    for idx in sample_idx:
        #_, gene_idx, _ = np.intersect1d(liger_object.adata_list[idx].var.index, genes_shared, return_indices=True)
        _, gene_idx, _ = np.intersect1d(liger_object.adata_list[idx].raw.var.index, genes_shared, return_indices=True)
        gene_idx = np.sort(gene_idx)  # make sure data are extracted in the original order
        #gene_name = liger_object.adata_list[idx].var.index[gene_idx]
        gene_name = liger_object.adata_list[idx].raw.var.index[gene_idx]
        raw_data = liger_object.adata_list[idx].raw.X[:, gene_idx]
        norm_data = sp_normalize(raw_data, axis=1, norm='l1')
        feature_matrix.append(norm_data)
        clusters = np.concatenate((clusters, liger_object.adata_list[idx].obs['cluster']), axis=None)
        cell_source = np.concatenate((cell_source, np.repeat(liger_object.adata_list[idx].uns['sample_name'],
                                                             liger_object.adata_list[idx].shape[0])), axis=None)
    feature_matrix = vstack(feature_matrix)

    ### Perform wilcoxon test
    if compare_method == 'clusters':  # compare between clusters across datasets
        num_rows = feature_matrix.shape[1]
        if num_rows > 100000:
            print('Calculating Large-scale Input...')
        results = _wilcoxon(np.log(feature_matrix.toarray() + 1e-10), clusters)
        gene_label = np.tile(gene_name, len(np.unique(clusters)))
        cluster_label = np.repeat(np.unique(clusters), len(gene_name))
    elif compare_method == 'datasets':  # compare between datasets within each cluster
        groups = np.unique(clusters)
        cluster_label = np.array([])
        gene_label = np.array([])
        data = []
        for group in groups:
            sub_idx = clusters == group
            sub_clusters = clusters[sub_idx]
            sub_cell_source = cell_source[sub_idx]
            sub_label = np.array(
                [str(int(sub_clusters[i])) + '-' + sub_cell_source[i] for i in range(len(sub_clusters))])
            sub_matrix = feature_matrix[sub_idx, :]
            if len(np.unique(sub_cell_source)) == 1:
                print('Note: Skip Cluster {} since it has only ONE data source.'.format(group))
            else:
                data.append(_wilcoxon(np.log(sub_matrix.toarray() + 1e-10), sub_label))
                gene_label = np.concatenate((gene_label, np.tile(gene_name, len(np.unique(sub_label)))), axis=None)
                cluster_label = np.concatenate((cluster_label, np.repeat(np.unique(sub_label), len(gene_name))),
                                               axis=None)
        results = pd.concat(data, ignore_index=True)

    results.insert(0, 'group', cluster_label)
    results.insert(0, 'feature', gene_label)
    return results

def _wilcoxon(X, y, groups_use=None, verbose=True, lable_use=None):
    ### Check and possibly correct input values
    if X.shape[0] != len(y):
        raise ValueError('Number of columns of X does not match length of y')

    if groups_use is not None:
        idx_use = np.isin(groups_use, y)
        X = X[idx_use, :]
        y = y[idx_use]

    idx_use = pd.isnull(y)
    if np.sum(~idx_use) < len(y):
        X = X[~idx_use, :]
        y = y[~idx_use]
        if verbose:
            print('Removing NA values from labels')

    ### Compute primary statistics
    clusters, cluster_size = np.unique(y, return_counts=True)
    num_genes = X.shape[1]
    num_clusters = len(clusters)
    statistic = []
    pval = []
    padj = []
    sum_table = np.empty((num_clusters, num_genes), dtype=np.float64)
    # ranks = rankdata(X, axis=0)
    temp_z = []
    for i in range(num_genes):
        ranks, T = _rank(X[:, i])
        temp_pval = []
        for j in range(num_clusters):
            idx = y == clusters[j]
            sum_table[j, i] = np.sum(X[idx, i])
            u, z = _mannwhitneyu(ranks[idx], ranks[~idx], T)
            statistic.append(u)
            temp_pval.append(z)
            temp_z.append(z)
        pval.extend(temp_pval)
        padj.extend(smt.multipletests(temp_pval, method='fdr_bh')[1])
    pval = 2 * distributions.norm.sf(np.abs(temp_z))

    ### Auxiliary Statistics (AvgExpr, LFC, etc)
    cs = np.sum(sum_table, axis=0, dtype=np.float64)
    avg_expr = np.divide(sum_table, np.reshape(cluster_size, (cluster_size.shape[0], 1)), dtype=np.float64)
    lfc = np.ravel([avg_expr[i, :] - ((cs - sum_table[i, :]) / (len(y) - cluster_size[i]))
                    for i in range(num_clusters)])

    summary = {'avgExpr': np.ravel(avg_expr),
               'logFC': lfc,
               'statistic': np.asarray(statistic),
               'pval': np.asarray(pval),
               'padj': np.asarray(padj)}
    return pd.DataFrame(data=summary)

from numba import jit, njit

@njit
def _mannwhitneyu(X1, X2, T):
    """"""
    n1 = len(X1)
    n2 = len(X2)

    u1 = n1 * n2 + (n1 * (n1 + 1)) / 2.0 - np.sum(X1)
    u2 = n1 * n2 - u1

    sd = np.sqrt(T * n1 * n2 * (n1 + n2 + 1) / 12.0)

    meanrank = n1 * n2 / 2.0 + 0.5
    bigu = max(u1, u2)

    z = (bigu - meanrank) / sd
    # p = 2 * distributions.norm.sf(np.abs(z))
    u = min(u1, u2)
    return u, z


def _rank(arr):
    """
    Parameters
    ----------
    arr : numpy array
        array needs to be ranked

    Returns
    -------
    ranks : numpy array

    """
    sorter = np.argsort(arr)
    inv = np.empty(len(sorter), dtype=np.intp)
    inv[sorter] = np.arange(len(sorter), dtype=np.intp)
    arr = arr[sorter]

    obs = np.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv]
    count = np.r_[np.nonzero(obs)[0], len(obs)]

    ranks = .5 * (count[dense] + count[dense - 1] + 1)

    cnt = np.diff(count).astype(np.float64)
    size = np.float64(arr.size)

    T = 1.0 - (cnt ** 3 - cnt).sum() / (size ** 3 - size)

    return ranks, T


"""
    for i in range(num_clusters):
        idx = y == clusters[i]
        ranks = _rank(X[])
        sum_table[i, :] = np.sum(X[idx, :], axis=0, dtype=np.float64)
        temp_pval = []
        for j in range(num_genes):
            try:
                #u, p = mannwhitneyu(X[idx, j], X[~idx, j], alternative='two-sided')
                u, p = _mannwhitneyu(ranks[idx, j], ranks[~idx, j], alternative='two-sided')
            except ValueError:
                u, p = -1.0, 1.0
            statistic.append(u)
            temp_pval.append(p)
        pval.extend(temp_pval)
        padj.extend(smt.multipletests(temp_pval, method='fdr_bh')[1])
"""

"""
def wilcoxon(X, y, groups_use=None, verbose=True, lable_use=None):
    #helper function for wilcoxon tests on single-cell datasets (cell by gene matrix)
    ### Check and possibly correct input values
    if X.shape[0] != len(y):
        raise ValueError('Number of columns of X does not match length of y')

    if groups_use is not None:
        idx_use = np.isin(groups_use, y)
        X = X[idx_use, :]
        y = y[idx_use]

    idx_use = pd.isnull(y)
    if np.sum(~idx_use) < len(y):
        X = X[~idx_use, :]
        y = y[~idx_use]
        if verbose:
            print('Removing NA values from labels')

    ### Compute primary statistics
    clusters, cluster_size = np.unique(y, return_counts=True)
    num_genes = X.shape[1]
    num_clusters = len(clusters)
    statistic = []
    pval = []
    padj = []
    sum_table = np.zeros((num_clusters, num_genes), dtype=np.float64)

    for i in range(num_clusters):
        idx = y == clusters[i]
        sum_table[i, :] = np.sum(X[idx, :], axis=0, dtype=np.float64)
        temp_pval = []
        for j in range(num_genes):
            try:
                u, p = mannwhitneyu(X[idx, j], X[~idx, j], alternative='two-sided')
            except ValueError:
                u, p = -1.0, 1.0
            statistic.append(u)
            temp_pval.append(p)
        pval.extend(temp_pval)
        padj.extend(smt.multipletests(temp_pval, method='fdr_bh')[1])

    ### Auxiliary Statistics (AvgExpr, LFC, etc)
    cs = np.sum(sum_table, axis=0, dtype=np.float64)
    avg_expr = np.divide(sum_table, np.reshape(cluster_size, (cluster_size.shape[0], 1)), dtype=np.float64)
    lfc = np.ravel([avg_expr[i, :] - ((cs - sum_table[i, :]) / (len(y) - cluster_size[i])) 
                      for i in range(num_clusters)])

    summary = {'avgExpr': np.ravel(avg_expr),
               'logFC': lfc,
               'statistic': np.asarray(statistic),
               'pval': np.asarray(pval),
               'padj': np.asarray(padj)}
    return pd.DataFrame(data=summary)
"""



import numpy as np
from scipy import interpolate
from scipy.stats.mstats import mquantiles

from pyliger.clustering._utilities import refine_clusts

    
def quantile_norm(liger_object, 
                  quantiles=50,
                  ref_dataset=None,
                  min_cells=20,
                  dims_use=None,
                  do_center=False,
                  max_sample=1000,
                  num_trees=None,
                  refine_knn=True,
                  knn_k=20,
                  use_ann=False,
                  rand_seed=1):
    """Quantile align (normalize) factor loadings
    
    This process builds a shared factor neighborhood graph to jointly cluster cells, then quantile
    normalizes corresponding clusters.

    The first step, building the shared factor neighborhood graph, and
    produces a graph representation where edge weights between cells (across all datasets)
    correspond to their similarity in the shared factor neighborhood space. An important parameter
    here is knn_k, the number of neighbors used to build the shared factor space. 
    
    Next we perform quantile alignment for each dataset, factor, and cluster (by
    stretching/compressing datasets' quantiles to better match those of the reference dataset). These
    aligned factor loadings are combined into a single matrix and returned as H_norm.
    
    Parameters
    ----------
    liger_object : liger
        Should run optimizeALS before calling.
    quantiles : int, optional
        Number of quantiles to use for quantile normalization (the default is 50).
    ref_dataset : str, optional
        Name of dataset to use as a "reference" for normalization. By default,
        the dataset with the largest number of cells is used (the default is None).
    min_cells : int, optional
        Minimum number of cells to consider a cluster shared across datasets
        (the default is 20).
    dims_use : list, optional
        Indices of factors to use for shared nearest factor determination
        (the default is list(range(liger_object.adata_list[0].varm['H'].shape[1]))).
    do_center : bool, optional
        Centers the data when scaling factors (useful for less sparse modalities like
        methylation data) (the default is False).
    max_sample : int, optional
        Maximum number of cells used for quantile normalization of each cluster 
        and factor (the default is 1000).
    num_trees : int, optional
        The number of trees used by the approximate nearest neighbor search. 
        Larger number of trees give more precision but consume more memory.
        According to largeVis, they use 10 trees for datasets has less than 100,000
        observations, 20 trees for datasets has less than 1,000,000 observations,
        50 trees for datasets up to 5,000,000 and 100 trees otherwise (the default is None).
    refine_knn : bool, optional
        whether to increase robustness of cluster assignments using KNN graph
        (the default is True).
    knn_k : int, optional
        Number of nearest neighbors for within-dataset knn graph (the default is 20). 
    use_ann : bool, optional
        Whether to use approximate nearest neighbor (the default is False)
    rand_seed : int, optional
        Random seed to allow reproducible results (the default is 1).

    Returns
    -------
    liger_object : liger
        liger_object with 'H_norm' and 'clusters' attributes.

    Examples
    --------
    ligerex = quantile_norm(ligerex) # do basic quantile alignment
    ligerex = quantile_norm(ligerex, resolution = 1.2) # higher resolution for more clusters (note that SNF is conserved)
    ligerex = quantile_norm(ligerex, knn_k = 15, resolution = 1.2) # change knn_k for more fine-grained local clustering
    """
    
    np.random.seed(rand_seed)
    
    num_samples = len(liger_object.adata_list)

    # set reference dataset
    if ref_dataset is None:
        ns = [adata.shape[0] for adata in liger_object.adata_list]
        ref_dataset_idx = np.argmax(ns)
    else:
        for i in range(num_samples):
            if liger_object.adata_list[i].uns['sample_name'] == ref_dataset:
                ref_dataset_idx = i
                break

    # set indices of factors
    if dims_use is None:
        use_these_factors = list(range(liger_object.adata_list[0].obsm['H'].shape[1]))
    else:
        use_these_factors = dims_use
    
    # applied use_these_factors to Hs
    Hs = [adata.obsm['H'][:, use_these_factors] for adata in liger_object.adata_list]
    num_clusters = Hs[ref_dataset_idx].shape[1]

    ### Max factor assignment
    clusters = []
    for i in range(num_samples):
        # scale the H matrix by columns, equal to scale_columns_fast in Rcpp
        if do_center:
            scale_H = (Hs[i]-np.mean(Hs[i], axis=0))/np.std(Hs[i], axis=0, ddof=1)
        else:
            scale_H = Hs[i]/(np.sqrt(np.sum(np.square(Hs[i]), axis=0)/(Hs[i].shape[0]-1)))

        # get the index of maximum value for each cell, equal to max_factor in Rcpp
        clusts = np.argmax(scale_H, axis=1)
        
        # increase robustness of cluster assignments using knn graph
        if refine_knn:
            clusts = refine_clusts(Hs[i], clusts, knn_k, use_ann, num_trees)
        
        clusters.append(clusts)
        
        # assign clusters to corresponding adata
        liger_object.adata_list[i].obs['cluster'] = clusts

    
    # all H_matrix used for quantile alignment
    Hs = [np.copy(adata.obsm['H']) for adata in liger_object.adata_list]
    
    ### Perform quantile alignment
    for k in range(num_samples):
        for j in range(num_clusters):
            cells2 = clusters[k] == j
            cells1 = clusters[ref_dataset_idx] == j

            for i in use_these_factors:
                num_cells2 = np.sum(cells2)
                num_cells1 = np.sum(cells1)
                
                # skip clusters having too less cells
                if num_cells1 < min_cells or num_cells2 < min_cells:
                    continue
                
                if num_cells2 == 1:
                    Hs[k][cells2, i] = np.mean(Hs[ref_dataset_idx][cells1, i])
                    continue
                
                # maximum number of cells used for quantile normalization
                q2 = mquantiles(np.random.permutation(Hs[k][cells2, i])[0:min(num_cells2, max_sample)],
                                np.linspace(0, 1, num=quantiles + 1), alphap=1, betap=1)
                max_H = np.max(Hs[k][cells2, i])
                min_H = np.min(Hs[k][cells2, i])
                if q2[-1] < max_H:
                    q2[-1] = max_H
                if q2[0] > min_H:
                    q2[0] = min_H
                q1 = mquantiles(np.random.permutation(Hs[ref_dataset_idx][cells1, i])[0:min(num_cells1, max_sample)], np.linspace(0, 1, num=quantiles+1),
                                alphap=1, betap=1)

                if np.sum(q1) == 0 or np.sum(q2) == 0 or len(np.unique(q1)) < 2 or len(np.unique(q2)) < 2:
                    new_vals = np.repeat(0, num_cells2) 
                else:
                    # handle ties (zeros) in order to get consistent results with LIGER
                    q1 = _mean_ties(q2, q1)
                    warp_func = interpolate.interp1d(q2, q1)
                    new_vals = warp_func(Hs[k][cells2, i])
                Hs[k][cells2, i] = new_vals
        
        # assign H_norm to corresponding adata
        liger_object.adata_list[k].obsm['H_norm'] = Hs[k]         
    
    return None


def _mean_ties(x, y):
    """helper function to calculate the mean value of y where ties(zeros) occur in x"""
    idx_zeros = x == 0
    if np.sum(idx_zeros) > 0:
        y[idx_zeros] = np.mean(y[idx_zeros])

    return y


import h5sparse
import numpy as np
from pathlib import Path
from sklearn.preprocessing import normalize as sp_normalize

from ..pyliger import Liger
from .._utilities import _h5_idx_generator, _remove_missing_obs

from typing import Optional


PARRENT_PATH = Path(__file__).parent

def normalize(liger_object: Liger,
              remove_missing: bool = True,
              chunk_size: Optional[int] = 1000
              ) -> None:
    """Normalize raw datasets to row sums

    This function normalizes data to account for total gene expression across a cell.

    Parameters
    ----------
    liger_object : liger object
        liger object with raw_data.
    chunk_size :

    remove_missing :

    Returns
    -------
    liger_object : liger object
        liger object with norm_data.

    Examples
    --------
    >>> adata1 = AnnData(np.arange(12).reshape((4, 3)))
    >>> adata2 = AnnData(np.arange(12).reshape((4, 3)))
    >>> ligerex = create_liger([adata1, adata2])
    >>> ligerex = normalize(ligerex)
    """
    # Iterate through each sample
    for idx, adata in enumerate(liger_object.adata_list):

        # On-disk mode (set for online learning approach)
        if adata.isbacked:
            norm_sum, norm_sum_sq = _normalize_online(adata, chunk_size)

        # In-memory mode
        else:
            norm_data, norm_sum, norm_sum_sq = _normalize_matrix(adata, remove_missing)
            liger_object.adata_list[idx].layers['norm_data'] = norm_data

        # save row sum and sum of squares for further use
        liger_object.adata_list[idx].var['norm_sum'] = norm_sum
        liger_object.adata_list[idx].var['norm_sum_sq'] = norm_sum_sq
        liger_object.adata_list[idx].var['norm_mean'] = norm_sum / adata.shape[0]

    return None


def _normalize_online(adata, chunk_size):
    """"""
    norm_sum = np.zeros(adata.shape[1])
    norm_sum_sq = np.zeros(adata.shape[1])

    # create h5 file for each individual sample.
    file_path = './results/' + adata.uns['sample_name'] + '.hdf5'
    with h5sparse.File(file_path, 'r+') as f:
        for left, right in _h5_idx_generator(chunk_size, adata.shape[0]):
            # normalize data and store normalized data as sparse matrix in h5 file
            #norm_data = sp_normalize(adata.X[left:right, :], axis=1, norm='l1')
            norm_data = sp_normalize(f['raw_data'][left:right][:, ~adata.uns['idx_missing']], axis=1, norm='l1')
            if 'norm_data' not in f.keys():
                f.create_dataset('norm_data', data=norm_data, chunks=(chunk_size,), maxshape=(None,))
            else:
                f['norm_data'].append(norm_data)


            # calculate row sum and sum of squares using normalized data
            norm_sum = norm_sum + np.ravel(np.sum(norm_data, axis=0))
            norm_sum_sq = norm_sum_sq + np.ravel(np.sum(norm_data.power(2), axis=0))

    return norm_sum, norm_sum_sq


def _normalize_matrix(adata, remove_missing):
    """"""
    if remove_missing:
        adata = _remove_missing_obs(adata, slot_use='raw_data', use_rows=True)
    norm_data = sp_normalize(adata.X, axis=1, norm='l1')
    norm_sum = np.ravel(np.sum(norm_data, axis=0))
    norm_sum_sq = np.ravel(np.sum(norm_data.power(2), axis=0))

    return norm_data, norm_sum, norm_sum_sq


import numpy as np

from tqdm import tqdm
from ._utilities import nnlsm_blockpivot

def optimize_ALS(liger_object,
                 k,
                 value_lambda=5.0,
                 thresh=1e-6,
                 max_iters=30,
                 nrep=1,
                 H_init=None,
                 W_init=None,
                 V_init=None,
                 rand_seed=1,
                 print_obj=False):
    """ Perform iNMF on scaled datasets

    Perform integrative non-negative matrix factorization to return factorized H, W, and V matrices.
    It optimizes the iNMF objective function using block coordinate descent (alternating non-negative
    least squares), where the number of factors is set by k. TODO: include objective function equation here in documentation (using deqn)

    For each dataset, this factorization produces an H matrix (cells by k), a V matrix (k by genes),
    and a shared W matrix (k by genes). The H matrices represent the cell factor loadings.
    W is held consistent among all datasets, as it represents the shared components of the metagenes
    across datasets. The V matrices represent the dataset-specific components of the metagenes.

    Args:
        liger_object(liger):
            Should normalize, select genes, and scale before calling.
        k(int):
            Inner dimension of factorization (number of factors). Run suggestK to determine
            appropriate value; a general rule of thumb is that a higher k will be needed for datasets with
            more sub-structure.
        value_lambda(float): optional, (default 5.0)
            Regularization parameter. Larger values penalize dataset-specific effects more
            strongly (ie. alignment should increase as lambda increases). Run suggestLambda to determine
            most appropriate value for balancing dataset alignment and agreement.
        thresh(float): optional, (default 1e-6)
            Convergence threshold. Convergence occurs when |obj0-obj|/(mean(obj0,obj)) < thresh.
        max_iters(int): optional, (default 30)
            Maximum number of block coordinate descent iterations to perform
        nrep(int): optional, (default 1)
            Number of restarts to perform (iNMF objective function is non-convex, so taking the
            best objective from multiple successive initializations is recommended). For easier
            reproducibility, this increments the random seed by 1 for each consecutive restart, so future
            factorizations of the same dataset can be run with one rep if necessary.
        H_init(): optional, (default None)
            Initial values to use for H matrices.
        W_init(): optional, (default None)
            Initial values to use for W matrix.

        V_init(): optional, (default None)
            Initial values to use for V matrices.
        rand_seed(seed): optional, (default 1)
            Random seed to allow reproducible results
        print_obj(bool): optional, (default False)
            Print objective function values after convergence.

    Return:
        liger_object(liger):
            liger object with H, W, and V attributes.

    Usage:
        >>> adata1 = AnnData(np.arange(12).reshape((4, 3)))
        >>> adata2 = AnnData(np.arange(12).reshape((4, 3)))
        >>> ligerex = pyliger.create_liger([adata1, adata2])
        >>> ligerex = pyliger.normalize(ligerex)
        >>> ligerex = pyliger.select_genes(ligerex) # select genes
        >>> ligerex = pyliger.scale_not_center(ligerex)
        >>> ligerex = pyliger.optimize_ALS(ligerex, k = 20, value_lambda = 5, nrep = 3) # get factorization using three restarts and 20 factors
    """
    ### 0. Extract required information
    # prepare basic dataset profiles
    N = liger_object.num_samples  # number of total input hdf5 files
    ns = [adata.shape[0] for adata in liger_object.adata_list]  # number of cells in each hdf5 files
    num_genes = len(liger_object.var_genes)  # number of variable genes
    X = [adata.layers['scale_data'].toarray() for adata in liger_object.adata_list]

    if k >= np.min(ns):
        raise ValueError('Select k lower than the number of cells in smallest dataset: {}'.format(np.min(ns)))

    if k >= num_genes:
        raise ValueError('Select k lower than the number of variable genes: {}'.format(len(liger_object.var_genes)))

    best_obj = np.Inf

    for j in range(nrep):
        np.random.seed(seed=rand_seed + j - 1)

        ### 1. Initialization (W, V_i, H_i)
        W = np.abs(np.random.uniform(0, 2, (k, num_genes)))
        V = [np.abs(np.random.uniform(0, 2, (k, num_genes))) for i in range(N)]
        H = [np.abs(np.random.uniform(0, 2, (ns[i], k))) for i in range(N)]

        if W_init is not None:
            W = W_init

        if V_init is not None:
            V = V_init

        if H_init is not None:
            H = H_init

        delta = 1
        sqrt_lambda = np.sqrt(value_lambda)

        # Initial training obj
        obj_train_approximation = 0
        obj_train_penalty = 0
        for i in range(N):
            obj_train_approximation += np.linalg.norm(X[i] - H[i] @ (W + V[i])) ** 2
            obj_train_penalty += np.linalg.norm(H[i] @ V[i]) ** 2

        obj0 = obj_train_approximation + value_lambda * obj_train_penalty

        ### 2. Iteration starts here
        for iter in tqdm(range(max_iters)):
            if delta > thresh:
                ## 1) update H matrix
                for i in range(N):
                    H[i] = nnlsm_blockpivot(A=np.hstack(((W + V[i]), sqrt_lambda * V[i])).transpose(),
                                            B=np.hstack((X[i], np.zeros((ns[i], num_genes)))).transpose())[0].transpose()

                ## 2) update V matrix
                for i in range(N):
                    V[i] = nnlsm_blockpivot(A=np.vstack((H[i], sqrt_lambda * H[i])),
                                            B=np.vstack(((X[i] - H[i] @ W), np.zeros((ns[i], num_genes)))))[0]

                ## 3) update W matrix
                W = nnlsm_blockpivot(A=np.vstack(H), B=np.vstack([(X[i] - H[i] @ V[i]) for i in range(N)]))[0]

                obj_train_prev = obj0
                obj_train_approximation = 0
                obj_train_penalty = 0
                for i in range(N):
                    obj_train_approximation += np.linalg.norm(X[i] - H[i] @ (W + V[i])) ** 2
                    obj_train_penalty += np.linalg.norm(H[i] @ V[i]) ** 2
                obj0 = obj_train_approximation + value_lambda * obj_train_penalty
                delta = np.absolute(obj_train_prev - obj0) / ((obj_train_prev + obj0) / 2)
            else:
                continue

        if obj0 < best_obj:
            final_W = W
            final_H = H
            final_V = V
            best_obj = obj0
            best_seed = rand_seed + i - 1

        if print_obj:
            print('Objective: {}'.format(best_obj))

    #liger_object.W = final_W.transpose()

    ### 3. Save results into the liger_object
    for i in range(N):
        liger_object.adata_list[i].obsm['H'] = final_H[i]
        liger_object.adata_list[i].varm['W'] = final_W.transpose()
        liger_object.adata_list[i].varm['V'] = final_V[i].transpose()
        #idx = liger_object.adata_list[i].uns['var_gene_idx']
        #shape = liger_object.adata_list[i].shape
        #save_W = np.zeros((shape[1], k))
        #save_W[idx, :] = final_W.transpose()
        #save_V = np.zeros((shape[1], k))
        #save_V[idx, :] = final_V[i].transpose()
        #liger_object.adata_list[i].obsm['H'] = final_H[i]
        #liger_object.adata_list[i].varm['W'] = save_W
        #liger_object.adata_list[i].varm['V'] = save_V

    return None



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from plotnine import *

from ._utilities import get_gene_values


def plot_spatial(liger_object,
                 gene,
                 log2scale=True,
                 image_key='lowres',
                 alpha=0.5,
                 crop=True,
                 zero_color='#F5F5F5',
                 return_plots=False):
    """

    :param liger_object:
    :param image_res:
    :param crop:
    :param zero_color:
    :param return_plots:
    :return:
    """
    figs = []
    for adata in liger_object.adata_list:
        gene_vals = get_gene_values(liger_object, gene, log2scale=log2scale)
        figs = _plot_spatial_adata(adata, gene, image_key, gene_vals)


    if return_plots:
        return figs
    else:
        return None


def _plot_spatial_adata(adata,
                        gene,
                        image_key,
                        gene_vals):
    """"""
    res_dict = {'lowres': 'tissue_lowres_scalef',
                'hires': 'tissue_hires_scalef'}

    ### 1. Extract image information
    pxl_in_fullres = adata.obsm['pxl_in_fullres']
    scale = adata.uns['image']['scale_factors'][res_dict[image_key]]
    height, width, _ = adata.uns['image'][image_key].shape

    pt_size = adata.uns['image']['scale_factors']['spot_diameter_fullres'] * scale * 1/32
    df = _coordinate_transfer(pxl_in_fullres, scale, height)
    df['gene'] = gene_vals

    option = 'plasma'
    zero_color = '#F5F5F5'

    ### 2. Plotting
    ggp = (ggplot(data=df, mapping=aes(x='x', y='y', color='gene')) +
           geom_point(size=pt_size) +
           xlim(0, width) +
           ylim(0, height) +
           labs(colour=gene) +
           theme_classic(12) +
           theme(axis_line=element_blank(), axis_text_x=element_blank(),
                 axis_text_y=element_blank(), axis_ticks=element_blank(),
                 axis_title_x=element_blank(), axis_title_y=element_blank()) +
           scale_color_cmap(cmap_name=option, na_value=zero_color)
           )

    # attach spatial image
    fig = ggp.draw(return_ggplot=False, show=False)  # get matplotlib Figure object
    plt.close()
    ax = fig.get_axes()[0]
    ax.imshow(adata.uns['image'][image_key], extent=[0, width, 0, height])  # add spatial image layer

    return fig


def _coordinate_transfer(pxl_in_fullres, scale, height):
    """ Convert jpg coordinate into xy for plotting"""
    coords = pxl_in_fullres * scale

    coords[:, 0] = height - coords[:, 0]
    coords[:, [1, 0]] = coords[:, [0, 1]]

    df = pd.DataFrame(coords)
    df.columns = ['x', 'y']

    return df















import os
import mygene
import numpy as np

from pathlib import Path
#from __future__ import print_function

from goatools.base import download_go_basic_obo, download_ncbi_associations
from goatools.obo_parser import GODag
from goatools.anno.genetogo_reader import Gene2GoReader
from goatools.goea.go_enrichment_ns import GOEnrichmentStudyNS


def run_GO_analysis(gene_list,
                    background,
                    data_source,
                    result_path= None,
                    alpha=0.05,
                    methods=['fdr_bh']):
    """
    Wrapper function to run GOATOOLS
    :param gene_list:
    :param background:
    :param data_source:
    alpha: significance cut-off, default 0.5
    methods: defult multipletest correction method
    :return:
    """
    ### 0. Preprocessing parameters
    # determine data source
    if data_source == 'human':
        taxids = [9606]
    elif data_source == 'mouse':
        taxids = [10090]
    else:
        print('Data source can only be either human or mouse.')

    if result_path is None:
        obo = Path('./results', 'go-basic.obo')
        gene2go = Path('./results', 'gene2go')
    else:
        obo = Path(result_path, 'go-basic.obo')
        gene2go = Path(result_path, 'gene2go')

    ### 1. Download Ontologies and Associations
    obo_fname, fin_gene2go = _download_go(obo, gene2go)

    ### 2a. Load Ontologies and Associations
    obodag, ns2assoc = _load_go(obo_fname, fin_gene2go, taxids)

    ### 2b. Process background gene set
    if type(background[0]) == int:
        background_ids = background
    elif type(background[0]) == str:
        background_ids = _symbols_to_ids(background, taxids[0])

    ### 3. Initialize a GOEA object
    goeaobj = GOEnrichmentStudyNS(
        background_ids,  # List of mouse protein-coding genes
        ns2assoc,  # geneid/GO associations
        obodag,  # Ontologies
        propagate_counts=False,
        alpha=alpha,  # default significance cut-off
        methods=methods)  # defult multipletest correction method

    ### 4. Load study genes
    if type(gene_list[0]) == int:
        geneids_study = gene_list
    elif type(gene_list[0]) == str:
        geneids_study = _symbols_to_ids(gene_list, taxids[0])

    ### 5. Run GOEA
    goea_results_all = goeaobj.run_study(geneids_study)
    #goea_results_sig = [r for r in goea_results_all if r.p_fdr_bh < 0.05]

    return goea_results_all


def _symbols_to_ids(symbols, species):
    """Convert gene symbols to gene ids using mygene module"""
    mg = mygene.MyGeneInfo()
    mg_df = mg.querymany(symbols, scopes='symbol', species=species, as_dataframe=True)
    ids = mg_df['_id'].dropna()
    ids = ids[~ids.str.startswith('ENSG')]
    ids = ids.astype(int)
    ids = list(ids)

    return ids


def _download_go(obo, gene2go):
    """helper function to download Ontologies and Associations"""
    # Ontologies
    obo_fname = download_go_basic_obo(obo=obo)
    # Associations
    fin_gene2go = download_ncbi_associations(gene2go=gene2go)

    return obo_fname, fin_gene2go


def _load_go(obo_fname, fin_gene2go, taxids):
    """helper function to load Ontologies and Associations"""
    # Ontologies
    obodag = GODag(obo_fname)

    # Associaions
    # Read NCBI's gene2go. Store annotations in a list of namedtuples
    objanno = Gene2GoReader(fin_gene2go, taxids=taxids)

    # Get namespace2association where:
    #    namespace is:
    #        BP: biological_process
    #        MF: molecular_function
    #        CC: cellular_component
    #    assocation is a dict:
    #        key: NCBI GeneID
    #        value: A set of GO IDs associated with that gene
    ns2assoc = objanno.get_ns2assc()

    return obodag, ns2assoc


import os
import logging
import h5sparse
import numpy as np
from anndata import AnnData
from scipy.sparse import csr_matrix

from ..pyliger import Liger
from .._utilities import _remove_missing_obs, _h5_idx_generator, _merge_sparse_data_all

from typing import List, Optional


def create_liger(adata_list: List,
                 make_sparse: bool = True,
                 take_gene_union: bool = False,
                 remove_missing: bool = True,
                 chunk_size: Optional[int] = 1000
                 ) -> Liger:
    """Create a liger object.

    This function initializes a liger object with the raw data passed in. It requires a list of
    expression (or another single-cell modality) matrices (cell by gene) for at least two datasets.
    By default, it converts all passed data into Compressed Sparse Row matrix (CSR matrix) to reduce
    object size. It initializes cell_data with nUMI and nGene calculated for every cell.

    Parameters
    ----------
    adata_list : list
        List of AnnData objects which store expression matrices (cell by gene).
        Should be named by dataset.
    make_sparse : bool, optional
        Whether to convert raw_data into sparse matrices (the default is True).
    take_gene_union : bool, optional
        Whether to fill out raw_data matrices with union of genes across all
        datasets (filling in 0 for missing data) (requires make_sparse=True)
        (the default is False).
    remove_missing : bool, optional
        Whether to remove cells not expressing any measured genes, and genes not
        expressed in any cells (if take_gene_union=True, removes only genes not
        expressed in any dataset) (the default is True).
    chunk_size : int, optional


    Returns
    -------
    liger_object : liger object
        object with raw_data slot set.
    TODO: update the docstring for returns

    Examples
    --------
    >>> adata1 = AnnData(np.arange(12).reshape((4, 3)))
    >>> adata2 = AnnData(np.arange(12).reshape((4, 3)))
    >>> ligerex = pyliger.create_liger([adata1, adata2])

    """
    # On-disk mode (set for online learning approach)
    if adata_list[0].isbacked or 'backed_path' in adata_list[0].uns_keys():
        processed_list = []
        for adata in adata_list:
            processed_list.append(_initialization_online(adata, chunk_size, remove_missing))

        liger_object = Liger(processed_list)

    # In-memory mode
    else:
        liger_object = _create_liger_matrix(adata_list, make_sparse, take_gene_union, remove_missing)

    return liger_object


def _initialization_online(adata, chunk_size, remove_missing):
    """"""

    # calculate row sum and sum of squares using raw data
    gene_sum = np.zeros(adata.shape[1])
    gene_sum_sq = np.zeros(adata.shape[1])
    nUMI = np.zeros(adata.shape[0])
    nGene = np.zeros(adata.shape[0])

    file_path = './results/' + adata.uns['sample_name'] + '.hdf5'
    with h5sparse.File(file_path, 'r') as f:
        for left, right in _h5_idx_generator(chunk_size, adata.shape[0]):
            raw_data = csr_matrix(f['raw_data'][left:right])
            gene_sum += np.ravel(np.sum(raw_data, axis=0))
            gene_sum_sq += np.ravel(np.sum(raw_data.power(2), axis=0))
            nUMI[left:right] = np.ravel(np.sum(raw_data, axis=1))
            nGene[left:right] = raw_data.getnnz(axis=1)

    # save AnnData
    file_path = './results/' + adata.uns['sample_name'] + '.h5ad'
    if remove_missing:
        idx_missing = gene_sum == 0
    else:
        idx_missing = np.repeat(False, adata.shape[1])

    if np.sum(idx_missing) > 0:
        print('Removing {} genes not expressing in {}.'.format(np.sum(idx_missing), adata.uns['sample_name']))
    #processed_adata = adata[:, ~idx_missing].copy(file_path)
    #processed_adata.var['gene_sum'] = gene_sum[~idx_missing]
    #processed_adata.var['gene_sum_sq'] = gene_sum_sq[~idx_missing]
    #processed_adata.obs['nUMI'] = nUMI
    #processed_adata.obs['nGene'] = nGene
    adata = adata[:, ~idx_missing].copy(file_path)
    adata.var['gene_sum'] = gene_sum[~idx_missing]
    adata.var['gene_sum_sq'] = gene_sum_sq[~idx_missing]
    adata.obs['nUMI'] = nUMI
    adata.obs['nGene'] = nGene
    adata.uns['idx_missing'] = idx_missing

    return adata


def _create_liger_matrix(adata_list, make_sparse, take_gene_union, remove_missing):
    """"""
    from scipy.sparse import csr_matrix

    num_samples = len(adata_list)

    # Make matrix sparse
    if make_sparse:
        for idx, adata in enumerate(adata_list):
            # force raw data to be csr matrix
            adata_list[idx].X = csr_matrix(adata_list[idx].X, dtype=np.int)
            # check if dimnames exist
            if not adata.obs.index.name or not adata.var.index.name:
                raise ValueError('Raw data must have both row (cell) and column (gene) names.')
            # check whether cell name is unique or not
            if not adata.obs.index.is_unique and adata.X.shape[0] > 1:
                raise ValueError('At least one cell name is repeated across datasets; '
                                 'please make sure all cell names are unique.')

    # Take gene union (requires make_sparse=True)
    if take_gene_union and make_sparse:
        merged_data = _merge_sparse_data_all(adata_list)
        if remove_missing:
            missing_genes = np.ravel(np.sum(merged_data.X, axis=0)) == 0
            if np.sum(missing_genes) > 0:
                print('Removing {} genes not expressed in any cells across merged datasets.'.format(
                    np.sum(missing_genes)))
                # show gene name when the total of missing genes is less than 25
                if np.sum(missing_genes) < 25:
                    print(merged_data.var.index[missing_genes])
                # save data after removing missing genes
                merged_data = merged_data[:, ~missing_genes].copy()
        # fill out raw_data matrices with union of genes across all datasets
        for i in range(num_samples):
            adata_list[i] = merged_data[merged_data.obs.index == adata_list[i].index, :].copy()

    # Remove missing cells
    for idx, adata in enumerate(adata_list):
        if remove_missing:
            adata_list[idx] = _remove_missing_obs(adata, use_rows=True)
            # remove missing genes if not already merged
            if not take_gene_union:
                adata_list[idx] = _remove_missing_obs(adata, use_rows=False)

    # Create liger object based on raw data list
    liger_object = Liger(adata_list)

    # Initialize cell_data for liger_object with nUMI, nGene, and dataset
    for idx, adata in enumerate(adata_list):
        liger_object.adata_list[idx].var['gene_sum'] = np.ravel(np.sum(adata.X, axis=0))
        liger_object.adata_list[idx].var['gene_sum_sq'] = np.ravel(np.sum(adata.X.power(2), axis=0))
        liger_object.adata_list[idx].var['nCell'] = adata.X.getnnz(axis=0)
        liger_object.adata_list[idx].obs['nUMI'] = np.ravel(np.sum(adata.X, axis=1))
        liger_object.adata_list[idx].obs['nGene'] = adata.X.getnnz(axis=1)
        liger_object.adata_list[idx].obs['dataset'] = np.repeat(adata.uns['sample_name'], adata.shape[0])

    return liger_object





from ._iNMF_HALS import iNMF_HALS
from ._iNMF_ANLS import optimize_ALS
from ._online_iNMF import online_iNMF


#######################################################################################
#### Factorization TODO


# Perform factorization for new value of k
def optimizeNewK(liger_object, k_new, value_lambda=None, thresh=1e-4, max_iters=100,
                 rand_seed=1):
    pass


# Perform factorization for new data
def optimizeNewData(liger_object, new_data, which_datasets, add_to_existing=True,
                    value_lambda=None, thresh=1e-4, max_iters=100):
    pass


# Perform factorization for subset of data
def optimizeSubset(liger_object, cell_subset=None, cluster_subset=None,
                   value_lambda=None, thresh=1e-4, max_iters=100, datasets_scale=None):
    pass


# Perform factorization for new lambda value
def optimizeNewLambda(liger_object, new_lambda, thresh=1e-4, max_iters=100, rand_seed=1):
    pass


# Visually suggest appropriate lambda value
def suggestLambda(liger_object, k, lambda_test=None, rand_seed=1, num_cores=1,
                  thresh=1e-4, max_iters=100, knn_k=20, k2=500, ref_dataset=None,
                  resolution=1, gen_new=False, nrep=1, return_data=False, return_raw=False):
    pass


# Visually suggest appropiate k value
# k_test range is set from 5 to 55
def suggestK(liger_object, k_test=range(5, 55, 5), value_lambda=5, thresh=1e-4, max_iters=100,
             num_cores=1, rand_seed=1, gen_new=False, nrep=1, plot_log2=True,
             return_data=False, return_raw=False):
    pass



import numpy as np
import pandas as pd


def calculate_qc(liger_object):
    for adata in liger_object.adata_list:
        pass
    return None


def calculate_mt_pct(liger_object,
                     data_source):
    if data_source == 'human':
        pattern = 'MT-'
    elif data_source == 'mouse':
        pattern = 'mt-'

    for idx, adata in enumerate(liger_object.adata_list):
        liger_object.adata_list[idx] = _cal_mt_adata(adata, pattern)
        liger_object.adata_list[idx].uns['mt_pct'] = np.sum(liger_object.adata_list[idx].var['mt']) / liger_object.adata_list[idx].shape[1]

    return None

def _cal_mt_adata(adata,
                  pattern):
    adata.var['mt'] = adata.var.index.str.startswith(pattern)

    return adata

def filtering(liger_object,
              datasets_use = None,
              min_cells = 0,
              max_cells = np.inf,
              min_genes = 0,
              max_genes = np.inf,
              min_counts = 0,
              max_counts = np.inf,
              combine=False):

    num_samples = liger_object.num_samples

    if datasets_use is None:
        datasets_use = list(range(num_samples))

    if combine:
        gene_union = pd.concat([adata.var for adata in liger_object.adata_list])
        pass
    else:
        for i in datasets_use:
            liger_object.adata_list[i] = _filter_adata(liger_object.adata_list[i],
                                                      min_cells, max_cells, min_genes, max_genes, min_counts, max_counts)

def _filter_union():

    pass

def _filter_adata(adata,
                  min_cells,
                  max_cells,
                  min_genes,
                  max_genes,
                  min_counts,
                  max_counts):

    gene_filter = (adata.obs['nGene'] >= min_genes) & (adata.obs['nGene'] <= max_genes)
    cell_filter = (adata.var['nCell'] >= min_cells) & (adata.var['nCell'] <= max_cells)
    count_filter = (adata.var['gene_sum'] >= min_counts) & (adata.var['gene_sum'] <= max_counts)

    num_cell_rm = np.sum(~gene_filter)
    num_gene_rm = np.sum(~cell_filter&~count_filter)
    if num_cell_rm > 0:
        print('Removing {} cells not passed the threshold.'.format(num_cell_rm))

    if num_gene_rm > 0:
        print('Removing {} genes not passed the threshold.'.format(num_gene_rm))

    return adata[gene_filter, cell_filter&count_filter].copy()



from ._louvain import louvain_cluster
from ._leiden import leiden_cluster


#####TODO
def imputeKNN(liger_object,
              reference,
              queries,
              knn_k=20,
              weight=True,
              norm=True,
              scale=False):
    """Impute the query cell expression matrix

    Impute query features from a reference dataset using KNN.

    Parameters
    ----------
    liger_object : liger obect
        object.
    reference : str
        Dataset containing values to impute into query dataset(s).
    queries : TYPE
        Dataset to be augmented by imputation. If not specified, will pass in all datasets.
    knn_k : int, optional
        The maximum number of nearest neighbors to search (the default is 20).
    weight : bool, optional
        Whether to use KNN distances as weight matrix (the default is False).
    norm : bool, optional
        Whether normalize the imputed data with default parameters (the default is True).
    scale : bool, optional
        Whether scale but not center the imputed data with default parameters (the default is True).

    Returns
    -------
    liger_object : liger obect
        object with raw data in raw.data slot replaced by imputed data (genes by cells)
        # TODO: may not be replaced?

    Examples
    --------
    >>>
    """
    pass

# Linking genes to putative regulatory elements
def linkGenesAndPeaks(gene_counts, peak_counts, path_to_coords, genes_list = None, dist = "spearman",
                      alpha = 0.05):
    pass

# Export predicted gene-pair interaction
def makeInteractTrack(corr_mat, genes_list, output_path, path_to_coords):
    pass

# Analyze biological interpretations of metagene
def runGSEA(liger_object, gene_sets = [], mat_w = True, mat_v = 0, custom_gene_sets = []):
    pass



import h5sparse
import numpy as np
from scipy.sparse import csr_matrix
from sklearn.utils.sparsefuncs import inplace_column_scale

from .._utilities import _h5_idx_generator

from typing import Optional


def scale_not_center(liger_object,
                     remove_missing=True,
                     chunk_size: Optional[int] = 1000
                     ) -> None:
    """Scale genes by root-mean-square across cells

    This function scales normalized gene expression data after variable genes have been selected.
    Note that the data is not mean-centered before scaling because expression values must remain
    positive (NMF only accepts positive values). It also removes cells which do not have any
    expression across the genes selected, by default.

    Parameters
    ----------
    liger_object : liger object
        Should call normalize and selectGenes before calling.
    remove_missing : bool, optional
        Whether to remove cells from scale_data with no gene expression
        (the default is True).
    chunk_size :

    Returns
    -------
    liger_object : liger object
        Object with scale_data layer.

    Examples
    --------
    >>> adata1 = AnnData(np.arange(12).reshape((4, 3)))
    >>> adata2 = AnnData(np.arange(12).reshape((4, 3)))
    >>> ligerex = create_liger([adata1, adata2])
    >>> ligerex = normalize(ligerex)
    >>> ligerex = select_genes(ligerex) # select genes
    >>> ligerex = scale_not_center(ligerex)
    """

    for idx, adata in enumerate(liger_object.adata_list):
        var_gene_idx = adata.var.index.isin(liger_object.var_genes)

        # On-disk mode (set for online learning approach)
        if adata.isbacked:
            liger_object.adata_list[idx] = _scale_online(adata, var_gene_idx, chunk_size)

        # In-memory mode
        else:
            liger_object.adata_list[idx] = _scale_matrix(adata, var_gene_idx)

    return None


def _scale_online(adata, var_gene_idx, chunk_size):
    file_path = './results/' + adata.uns['sample_name'] + '.hdf5'
    with h5sparse.File(file_path, 'r+') as f:
        for left, right in _h5_idx_generator(chunk_size, adata.shape[0]):
            scale_data = csr_matrix(f['norm_data'][left:right][:, var_gene_idx] /
                                    np.sqrt(adata.var['norm_sum_sq'][var_gene_idx].to_numpy() /
                                    (adata.shape[0] - 1)), dtype=np.float64)
            if 'scale_data' not in f.keys():
                f.create_dataset('scale_data', data=scale_data, chunks=(chunk_size,), maxshape=(None,))
            else:
                f['scale_data'].append(scale_data)

    # slice adata after keeping a raw version
    #adata.raw = adata
    file_name = './results/' + adata.uns['sample_name'] + '.h5ad'
    adata = adata[:, var_gene_idx].copy(filename=file_name)

    return adata


def _scale_matrix(adata, var_gene_idx):
    """"""
    # slice adata after keeping a raw version
    adata.raw = adata
    adata = adata[:, var_gene_idx].copy()

    # calculate scale data
    scale_data = adata.copy().layers['norm_data']
    scaler = 1 / np.sqrt(adata.var['norm_sum_sq'].to_numpy() / (adata.shape[0] - 1))
    inplace_column_scale(scale_data, scaler)
    adata.layers['scale_data'] = scale_data

    #adata.layers['scale_data'] = csr_matrix(
    #    adata.layers['norm_data'] / np.sqrt(adata.var['norm_sum_sq'].to_numpy() / (adata.shape[0] - 1)),
    #    dtype=np.float64)
    #scale_data = csr_matrix(
    #    selected_data / np.sqrt(np.sum(np.square(selected_data.toarray()), axis=0) / (selected_data.shape[0] - 1)),
    #    dtype=np.float64)

    #numerical_idx = np.nonzero(var_gene_idx)[0]
    #row_idx, col_idx = scale_data.nonzero()
    #col_idx = np.take(numerical_idx, col_idx)
    #adata.layers['scale_data'] = csr_matrix((scale_data.data, (row_idx, col_idx)),
    #                                        shape=sample_shape, dtype=np.float64)

    # liger_object.adata_list[i] = liger_object.adata_list[i][:, idx].copy()
    # temp_norm = liger_object.adata_list[i].layers['norm_data']
    # liger_object.adata_list[i].layers['scale_data'] = csr_matrix(temp_norm / np.sqrt(np.sum(np.square(temp_norm.toarray()), axis=0) / (temp_norm.shape[0] - 1)))

    # if remove_missing:
    #    liger_object = _remove_missing_obs(liger_object, slot_use='scale_data', use_rows=False)
    return adata



import umap
import umap.plot
import numpy as np
import pandas as pd


def run_umap(liger_object,
             use_raw=False,
             dims_use=None,
             k=2,
             distance="euclidean",
             n_neighbors=10,
             min_dist=0.1,
             rand_seed=42):
    """Perform UMAP dimensionality reduction

    Run UMAP on the normalized cell factors (or raw cell factors) to generate a 2D embedding for
    visualization (or general dimensionality reduction). Has option to run on subset of factors.

    Note that running multiple times will overwrite tsne_coords values. It is generally
    recommended to use this method for dimensionality reduction with extremely large datasets.

    Parameters
    ----------
    liger_object : liger
        Should run quantile_norm before calling with defaults.
    use_raw : bool, optional
        Whether to use un-aligned cell factor loadings (H matrices) (the default is False).
    dims_use : list
        Factors to use for computing tSNE embedding (the default is using all cells).
    k : int, optional
        Number of dimensions to reduce to (the default is 2).
    distance : str, optional
        Mtric used to measure distance in the input space. A wide variety of metrics are
        already coded, and a user defined function can be passed as long as it has been JITd by numba.
        (the default is "euclidean", alternatives: "cosine", "manhattan", "hamming").
    n_neighbors : int, optional
        Number of neighboring points used in local approximations of manifold
        structure. Larger values will result in more global structure being preserved at the loss of
        detailed local structure. In general this parameter should often be in the range 5 to 50, with
        a choice of 10 to 15 being a sensible default (the default is 10).
    min_dist : float, optional
        Controls how tightly the embedding is allowed compress points together. Larger
        values ensure embedded points are more evenly distributed, while smaller values allow the
        algorithm to optimise more accurately with regard to local structure. Sensible values are in
        the range 0.001 to 0.5, with 0.1 being a reasonable default (the default is 0.1).
    rand_seed : int, optional
        Random seed for reproducibility (the default is 42).

    Returns
    -------
    liger_object : liger
        object with tsne_coords attribute.

    Examples
    --------
    >>> ligerex = quantile_norm(ligerex) # generate H_norm by quantile aligning factor loadings
    >>> ligerex = run_umap(ligerex) # get tsne_coords for normalized data
    >>> ligerex = run_umap(ligerex, use_raw = T) # get tsne.coords for raw factor loadings
    """
    np.random.seed(rand_seed)

    ### Run umap
    if use_raw:
        raw_data = np.vstack([adata.obsm['H'] for adata in liger_object.adata_list])

        # if H_norm not set yet
        if 'H_norm' not in liger_object.adata_list[0].obs_keys():
            dims_use = list(range(raw_data.shape[1]))
        else:
            H_norm = np.vstack([adata.obsm['H_norm'] for adata in liger_object.adata_list])
            dims_use = list(range(H_norm.shape[1]))

        umap_coords = umap.UMAP(n_components=k, metric=distance,
                                n_neighbors=n_neighbors, min_dist=min_dist,
                                random_state=rand_seed).fit_transform(raw_data[:, dims_use])
    else:
        #if dims_use is None:
        #    H_norm = np.vstack([adata.obsm['H_norm'] for adata in liger_object.adata_list])
        #    dims_use = list(range(H_norm.shape[1]))

        # tsne_coords = umap.UMAP(n_components=k, metric=distance,
        #                        n_neighbors=n_neighbors, min_dist=min_dist,
        #                        random_state=rand_seed).fit(H_norm[dims_use, :])
        H_norm = np.vstack([adata.obsm['H_norm'] for adata in liger_object.adata_list])
        umap_coords = umap.UMAP(n_components=k, metric=distance,
                                n_neighbors=n_neighbors, min_dist=min_dist,
                                random_state=rand_seed).fit_transform(H_norm)

    ### Save umap results
    liger_object.save_obsm(umap_coords, 'umap_coords')

    liger_object.tsne_coords = pd.DataFrame(umap_coords, columns=['tsne1', 'tsne2'])
    return None

# Perform t-SNE dimensionality reduction
def runTSNE(liger_object,
            dims_use,
            use_raw = False,
            use_pca = False,
            perplexity = 30,
            theta = 0.5,
            method = "Rtsne",
            fitsne_path = None,
            rand_seed = 42):
    #dims_use = range(1, len(liger_object.H_norm))
    pass


import numpy as np



def calc_dataset_specificity(liger_object, dataset1 = None, dataset2 = None, do_plot = True):
    """Calculate a dataset-specificity score for each factor

    This score represents the relative magnitude of the dataset-specific components of each factor's
    gene loadings compared to the shared components for two datasets. First, for each dataset we
    calculate the norm of the sum of each factor's shared loadings (W) and dataset-specific loadings
    (V). We then determine the ratio of these two values and subtract from 1... TODO: finish
    description.

    Parameters
    ----------
        liger_object:
        dataset1:
        dataset2:
        do_plot:

    Returns
    -------

    """
    ### Extract values for use
    W = liger_object.W
    V1 = liger_object.V[dataset1]
    V2 = liger_object.V[dataset2]

    ### Calculation
    pct1 = np.linalg.norm((V1 + W), axis=0)
    pct2 = np.linalg.norm((V2 + W), axis=0)

#    if do_plot:

    return [pct1, pct2, 100*(1-(pct1/pct2))]

# Calculate agreement metric
def calcAgreement(liger_object, dr_method = "NMF", ndims = 40, k = 15, use_aligned = True,
                  rand_seed = 42, by_dataset = False):
    pass

# Calculate alignment metric
def calcAlignment(liger_object, k = None, rand_seed = 1, cells_use = None, cells_comp = None,
                  clusters_use = None, by_cell = False, by_dataset = False):
    pass

# Calculate alignment for each cluster
def calcAlignmentPerCluster(liger_object, rand_seed = 1, k = None, by_dataset = False):
    pass

# Calculate adjusted Rand index
def calcARI(liger_object, clusters_compare):
    pass

# Calculate purity
def calcPurity(liger_object, classes_compare):
    pass

# Calculate proportion mitochondrial contribution
def getProportionMito(liger_object, use_norm = False):
    pass



import h5sparse
import numpy as np

from tqdm import tqdm
from scipy.sparse import vstack
from ..preprocessing._initialization import _initialization_online
from ..preprocessing._normalization import _normalize_online
from ..preprocessing._scale import _scale_online
from ._utilities import nnlsm_blockpivot, _init_W, _init_V_online, _update_W_HALS, _update_V_HALS, nonneg
from .._utilities import _h5_idx_generator

#from memory_profiler import profile
#@profile
def online_iNMF(liger_object,
                X_new = None,
                projection = False,
                W_init = None,
                V_init = None,
                H_init = None,
                A_init = None,
                B_init = None,
                k = 20,
                value_lambda = 5.0,
                max_epochs = 5,
                miniBatch_max_iters = 1,
                miniBatch_size = 5000,
                h5_chunk_size = 1000,
                rand_seed = 1,
                verbose = True
                ) -> None:
    """Perform online iNMF on scaled datasets

    Perform online integrative non-negative matrix factorization to represent multiple single-cell datasets
    in terms of H, W, and V matrices. It optimizes the iNMF objective function using online learning (non-negative
    least squares for H matrix, hierarchical alternating least squares for W and V matrices), where the
    number of factors is set by k. The function allows online learning in 3 scenarios: (1) fully observed datasets;
    (2) iterative refinement using continually arriving datasets; and (3) projection of new datasets without updating
    the existing factorization. All three scenarios require fixed memory independent of the number of cells.

    For each dataset, this factorization produces an H matrix (cells by k), a V matrix (k by genes),
    and a shared W matrix (k by genes). The H matrices represent the cell factor loadings.
    W is identical among all datasets, as it represents the shared components of the metagenes
    across datasets. The V matrices represent the dataset-specific components of the metagenes.

    Parameters
    ----------
        liger_object:
        X_new: AnnData object
        projection:
        W_init:
        V_init:
        H_init:
        A_init:
        B_init:
        k:
        value_lambda:
        max_epochs:
        miniBatch_max_iters:
        miniBatch_size:
        h5_chunk_size:
        rand_seed:
        verbose:

    Returns
    -------

    """
    matrices_init_dict = {'W_init': W_init,
                          'V_init': V_init,
                          'H_init': H_init,
                          'A_init': A_init,
                          'B_init': B_init}

    factorization_params = {'k': k,
                            'value_lambda': value_lambda,
                            'max_epochs': max_epochs,
                            'miniBatch_max_iters': miniBatch_max_iters,
                            'miniBatch_size': miniBatch_size,
                            'h5_chunk_size': h5_chunk_size,
                            'rand_seed': rand_seed}

    # Online Learning Scenario 1
    if X_new is None:
        if verbose:  # print start information
            print('Starting Online iNMF...')
        _online_iNMF_from_scratch(liger_object, verbose, matrices_init_dict, factorization_params)
    else:
        # check whether X_new needs to be processed
        for idx, adata in enumerate(X_new):
            if 'scale_data' not in adata.layers.keys():
                X_new[idx] = _preprocessing(adata, h5_chunk_size, liger_object.var_genes)

        # Online Learning Scenario 3
        if projection:
            if verbose:
                print("Metagene projection")
            _projection(liger_object, X_new, W_init, k, miniBatch_size)
        # Online Learning Scenario 2
        else:
            _online_iNMF_refine(liger_object, X_new, verbose, matrices_init_dict, factorization_params)

    return None


def _online_iNMF_from_scratch(liger_object,
                              verbose,
                              matrices_init_dict,
                              factorization_params):
    """ """
    ### 0. Extract required information
    # prepare basic dataset profiles
    num_files = liger_object.num_samples  # number of total input hdf5 files
    num_genes = len(liger_object.var_genes)  # number of variable genes
    num_cells = [adata.shape[0] for adata in liger_object.adata_list]  # number of cells in each hdf5 files
    prev_info = None
    Xs = [_get_scale_data(adata) for adata in liger_object.adata_list]

    #import scipy.io
    #from scipy.sparse import csr_matrix
    #Xs = [csr_matrix(scipy.io.mmread('/Users/lulu/Documents/GitHub/pyliger/results/stim.mtx').transpose()), csr_matrix(scipy.io.mmread('/Users/lulu/Documents/GitHub/pyliger/results/ctrl.mtx').transpose())]
    #Xs = [csr_matrix(scipy.io.mmread('/Users/lulu/Desktop/cells.mtx').transpose())]

    ### 1. Run Online_iNMF
    W, Vs, A, B = _online_iNMF_cal_W_V(Xs, num_genes, num_cells, verbose, prev_info, **matrices_init_dict, **factorization_params)

    Hs = _online_iNMF_cal_H(Xs, W, Vs, num_genes, num_cells, verbose,
                            factorization_params['value_lambda'], factorization_params['miniBatch_size'])


    ### 2. Sava results and close hdf5 files
    for file in Xs:  # close all files in the end
        if isinstance(file, h5sparse.h5sparse.File):
            file.close()

    for i in range(num_files):
        liger_object.adata_list[i].obsm['H'] = Hs[i].transpose()
        liger_object.adata_list[i].varm['W'] = W
        liger_object.adata_list[i].varm['V'] = Vs[i]
        liger_object.adata_list[i].varm['B'] = B[i]
        liger_object.adata_list[i].uns['A'] = A[i]

    return None


def _online_iNMF_refine(liger_object,
                        X_new,
                        verbose,
                        matrices_init_dict,
                        factorization_params):
    """ """
    if verbose:
        print('{} new datasets detected.'.format(len(X_new)))

    ### 0. Extract required information
    # prepare basic dataset profiles
    num_genes = len(liger_object.var_genes)  # number of variable genes

    if matrices_init_dict['W_init'] is None:
        matrices_init_dict['W_init'] = liger_object.W
    prev_info = {'A_prev': [adata.uns['A'] for adata in liger_object.adata_list],
                 'B_prev': [adata.varm['B'] for adata in liger_object.adata_list],
                 'V_prev': [adata.varm['V'] for adata in liger_object.adata_list]}

    for i, adata in enumerate(X_new):
        # open hdf5 files
        X = [_get_scale_data(adata)]

        #import scipy.io
        #from scipy.sparse import csr_matrix
        #X = [csr_matrix(scipy.io.mmread('/Users/lulu/Desktop/nuclei.mtx').transpose())]

        num_cells = [adata.shape[0]]

        # calculate W, V, H matrices for the new datasets
        W, Vs, A, B = _online_iNMF_cal_W_V(X, num_genes, num_cells, verbose, prev_info, **matrices_init_dict, **factorization_params)

        Hs = _online_iNMF_cal_H(X, W, Vs, num_genes, num_cells, verbose,
                                factorization_params['value_lambda'], factorization_params['miniBatch_size'])

        # Sava results and close hdf5 files
        if isinstance(X[0], h5sparse.h5sparse.File):
            X[0].close()
        X_new[i].obsm['H'] = Hs[0].transpose()
        X_new[i].varm['W'] = W
        X_new[i].varm['V'] = Vs[0]
        X_new[i].varm['B'] = B[0]
        X_new[i].uns['A'] = A[0]

        matrices_init_dict['W_init'] = W

    # update H matrics for existing datasets
    for i, adata in enumerate(liger_object.adata_list):
        X = [_get_scale_data(adata)]
        #X = [csr_matrix(scipy.io.mmread('/Users/lulu/Desktop/cells.mtx').transpose())]
        Vs = [adata.varm['V']]
        num_cells = [adata.shape[0]]
        Hs = _online_iNMF_cal_H(X, W, Vs, num_genes, num_cells, verbose,
                                factorization_params['value_lambda'], factorization_params['miniBatch_size'])

        if isinstance(X[0], h5sparse.h5sparse.File):
            X[0].close()
        liger_object.adata_list[i].obsm['H'] = Hs[0].transpose()
        liger_object.adata_list[i].varm['W'] = W

    for adata in X_new:
        liger_object.adata_list.append(adata)

    return None


def _projection(liger_object,
                X_new,
                W_init,
                k,
                miniBatch_size):
    """"""
    ### 0. Extract required information
    # prepare basic dataset profiles
    num_files = len(X_new)  # number of total input hdf5 files
    num_genes = len(liger_object.var_genes)  # number of variable genes
    num_cells = [adata.shape[0] for adata in X_new]  # number of cells in each hdf5 files

    ### 1. Initialization (W, V_i, A(HiHi^t), B(XiHit))
    W = liger_object.W if W_init is None else W_init

    # open hdf5 files
    Xs = [_get_scale_data(adata) for adata in X_new]

    #import scipy.io
    #from scipy.sparse import csr_matrix
    #Xs = [csr_matrix(scipy.io.mmread('/Users/lulu/Desktop/nuclei.mtx').transpose())]

    Hs = []
    Vs = []
    for i in range(num_files):
        num_batch = np.ceil(num_cells[i] / miniBatch_size).astype(int)
        H_miniBatch = []
        for batch in range(num_batch):
            if batch != num_batch:
                X_miniBatch = Xs[i]['scale_data'][batch*miniBatch_size:(batch+1)*miniBatch_size].transpose().toarray()
                #X_miniBatch = Xs[i][
                #              batch * miniBatch_size:(batch + 1) * miniBatch_size].transpose().toarray()
            else:
                X_miniBatch = Xs[i]['scale_data'][batch*miniBatch_size:num_cells[i]].transpose().toarray()
                #X_miniBatch = Xs[i][batch * miniBatch_size:num_cells[i]].transpose().toarray()

            H_miniBatch.append(nnlsm_blockpivot(A=W, B=X_miniBatch)[0])

        Hs.append(np.hstack(H_miniBatch))
        Vs.append(np.zeros((num_genes, k)))

    # Sava results and close hdf5 files
    for i in range(num_files):
        if isinstance(Xs[i], h5sparse.h5sparse.File):
            Xs[i].close()
        X_new[i].obsm['H'] = Hs[i].transpose()
        X_new[i].varm['W'] = W
        X_new[i].varm['V'] = Vs[i]

        # add object into liger object TODO: think of another way of appending adata into liger object
        liger_object.adata_list.append(X_new[i])

    return None


def _online_iNMF_cal_W_V(Xs,
                         num_genes,
                         num_cells,
                         verbose,
                         prev_info,
                         W_init,
                         V_init,
                         H_init,
                         A_init,
                         B_init,
                         k,
                         value_lambda,
                         max_epochs,
                         miniBatch_max_iters,
                         miniBatch_size,
                         h5_chunk_size,
                         rand_seed):
        """ """
        ### 1. Initialization (W, V_i, A(HiHi^t), B(XiHit))
        num_files = len(Xs)

        np.random.seed(seed=rand_seed)
        W = _init_W(num_genes, k, rand_seed) if W_init is None else W_init
        Vs = [_init_V_online(num_cells[i], k, Xs[i], h5_chunk_size, rand_seed) for i in range(num_files)] if V_init is None else V_init
        A = [np.zeros((k, k)) for i in range(num_files)] if A_init is None else A_init  # A = HiHi^t
        B = [np.zeros((num_genes, k)) for i in range(num_files)] if B_init is None else B_init  # B = XiHit

        # save information older than 2 epochs
        A_old = [np.zeros((k, k)) for i in range(num_files)]
        B_old = [np.zeros((num_genes, k)) for i in range(num_files)]

        ### 2. Iteration starts here
        # determine batch size for each dataset
        miniBatch_sizes = np.asarray(
            [np.round((num_cells[i] / np.sum(num_cells)) * miniBatch_size).astype(int) for i in range(num_files)])

        # determine total number of iterations
        total_iters = np.floor(num_cells[0] * max_epochs / miniBatch_sizes[0]).astype(int)

        # generate all slicing index for each iteration
        all_idx = [_generate_idx(total_iters, miniBatch_sizes[i], max_epochs, h5_chunk_size, num_cells[i]) for i in
                   range(num_files)]

        epoch = 0
        for iter in tqdm(range(total_iters)):
            # track epochs
            epoch_prev = epoch
            epoch = ((iter + 1) * miniBatch_sizes[0]) // num_cells[0]

            # cells from each dataset that are used in this iteration
            Xs_miniBatch = []
            for i in range(num_files):
                miniBatch_idx = all_idx[i][iter]
                X_miniBatch = vstack([Xs[i]['scale_data'][left:right] for left, right in miniBatch_idx])
                #X_miniBatch = vstack([Xs[i][left:right] for left, right in miniBatch_idx]) ###########
                Xs_miniBatch.append(X_miniBatch.transpose().toarray())

            ## 1) update H_i_minibatch by ANLS
            H_miniBatch = []
            for i in range(num_files):
                H_miniBatch.append(nnlsm_blockpivot(A=np.vstack(((W + Vs[i]), np.sqrt(value_lambda) * Vs[i])),
                                                    B=np.vstack(
                                                        (Xs_miniBatch[i], np.zeros((num_genes, miniBatch_sizes[i])))))[
                                       0])

            ## 2) updata A and B matrices
            for i in range(num_files):
                A[i], B[i], A_old[i], B_old[i] = _update_A_B(A[i], B[i], A_old[i], B_old[i], H_miniBatch[i],
                                                             Xs_miniBatch[i], miniBatch_sizes[i], iter, epoch,
                                                             epoch_prev)

            ## 3) update W, V_i by HALS
            iter_miniBatch = 1
            while iter_miniBatch <= miniBatch_max_iters:
                # update W
                if prev_info is None:
                    W = _update_W_HALS(A, B, W, Vs)
                else:
                    W = _update_W_HALS(prev_info['A_prev']+A, prev_info['B_prev']+B, W, prev_info['V_prev']+Vs)

                # update V_i
                Vs = _update_V_HALS(A, B, W, Vs, value_lambda)

                iter_miniBatch += 1

        return W, Vs, A, B


def _online_iNMF_cal_H(Xs,
                       W,
                       Vs,
                       num_genes,
                       num_cells,
                       verbose,
                       value_lambda,
                       miniBatch_size):
    """"""
    # Calculate metagene loadings (H metrics)
    num_files = len(Xs)
    if verbose:
        print('Calculate metagene loadings...')

    Hs = []
    for i in range(num_files):
        num_batch = np.ceil(num_cells[i] / miniBatch_size).astype(int)
        H_miniBatch = []
        for batch in range(num_batch):
            if batch != num_batch:
                X_miniBatch = Xs[i]['scale_data'][
                              batch * miniBatch_size:(batch + 1) * miniBatch_size].transpose().toarray()
                #X_miniBatch = Xs[i][batch * miniBatch_size:(batch + 1) * miniBatch_size].transpose().toarray()

            else:
                X_miniBatch = Xs[i]['scale_data'][batch * miniBatch_size:num_cells[i]].transpose().toarray()
                #X_miniBatch = Xs[i][batch * miniBatch_size:num_cells[i]].transpose().toarray()

            H_miniBatch.append(nnlsm_blockpivot(A=np.vstack(((W + Vs[i]), np.sqrt(value_lambda) * Vs[i])),
                                                B=np.vstack(
                                                    (X_miniBatch, np.zeros((num_genes, X_miniBatch.shape[1])))))[0])

        Hs.append(np.hstack(H_miniBatch))

    return Hs


def _generate_idx(num_iter, miniBatch_size, max_epochs, h5_chunk_size, num_cell):
    """helper function to generate miniBatch index for each iteration"""
    idx_dict = {}

    # permutate chunks for all+1 epochs (one more set for extreme case)
    all_idx = np.concatenate([_chunk_permutation(num_cell, h5_chunk_size) for j in range(max_epochs+1)])

    # assign chunks for each iteration
    for i in range(num_iter):
        total_sum = 0
        temp_list = []
        while total_sum < miniBatch_size:
            left, right = all_idx[0]
            diff = right - left
            missing = int(miniBatch_size - total_sum)
            if diff < missing:
                temp_list.append((left, right))
                all_idx = all_idx[1:]
                total_sum += diff
            elif diff == missing:
                temp_list.append((left, right))
                all_idx = all_idx[1:]
                total_sum += diff
            else:
                temp_list.append((left, left + missing))
                all_idx[0] = [left + missing, right]
                total_sum += missing
        idx_dict[i] = temp_list

    return idx_dict


def _chunk_permutation(num_cell, h5_chunk_size):
    """helper function to permutate chunks"""
    #all_chunks = np.random.permutation([(left, right) for left, right in _h5_idx_generator(h5_chunk_size, num_cell)])
    all_chunks = np.asarray([(left, right) for left, right in _h5_idx_generator(h5_chunk_size, num_cell)])

    return all_chunks


def _get_scale_data(adata):
    """helper function to open hdf5 files if AnnData is on-disk, otherwise get scale data from AnnData object"""
    if adata.isbacked:
        return h5sparse.File('./results/' + adata.uns['sample_name'] + '.hdf5', 'r')
    else:
        return adata.layers


def _update_A_B(A, B, A_old, B_old, H_miniBatch, X_miniBatch, miniBatch_size, iter, epoch, epoch_prev):
    """helper function to update A B matrices"""
    if iter == 0:
        scale_param = 0
    elif iter == 1:
        scale_param = 1 / miniBatch_size
    else:
        scale_param = (iter - 1) / iter

    if epoch > 0 and epoch - epoch_prev == 1:  # remove information older than 2 epochs
        A = A - A_old
        A_old = scale_param * A
        B = B - B_old
        B_old = scale_param * B
    else:  # otherwise scale the old information
        A_old = scale_param * A_old
        B_old = scale_param * B_old

    t_H_miniBatch = H_miniBatch.transpose()
    A = scale_param * A + (H_miniBatch @ t_H_miniBatch) / miniBatch_size
    B = scale_param * B + (X_miniBatch @ t_H_miniBatch) / miniBatch_size

    # replace the zero values in the diagonal of A with 1e-15
    A_diag = np.diagonal(A).copy()
    A_diag[A_diag == 0] = 1e-15
    np.fill_diagonal(A, A_diag)

    return A, B, A_old, B_old


def _preprocessing(adata, h5_chunk_size, var_genes):
    """helper function to preprocess AnnData object
    TODO: in the future, consider have preprocessing support for AnnData objects"""
    # initialization
    adata = _initialization_online(adata, h5_chunk_size, remove_missing=False)
    # normalization
    norm_sum, norm_sum_sq = _normalize_online(adata, h5_chunk_size)
    norm_sum = nonneg(norm_sum)  # to avoid zero-division
    norm_sum_sq = nonneg(norm_sum_sq)  # to avoid zero-division
    adata.var['norm_sum'] = norm_sum
    adata.var['norm_sum_sq'] = norm_sum_sq
    adata.var['norm_mean'] = norm_sum / adata.shape[0]
    # scale
    var_gene_idx = adata.var.index.isin(var_genes)
    adata = _scale_online(adata, var_gene_idx, h5_chunk_size)

    return adata
"""
    ### 1. Initialization (W, V_i, A(HiHi^t), B(XiHit))
    np.random.seed(seed=rand_seed)
    W = _init_W(num_genes, k, rand_seed) if W_init is None else W_init
    Vs = [_init_V_online(num_cells[i], k, Xs[i], h5_chunk_size, rand_seed) for i in range(num_files)] if V_init is None else V_init
    A = [np.zeros((k, k)) for i in range(num_files)] if A_init is None else A_init # A = HiHi^t
    B = [np.zeros((num_genes, k)) for i in range(num_files)] if B_init is None else B_init # B = XiHit

    # save information older than 2 epochs
    A_old = [np.zeros((k, k)) for i in range(num_files)]
    B_old = [np.zeros((num_genes, k)) for i in range(num_files)]

    ### 2. Iteration starts here
    # determine batch size for each dataset
    miniBatch_sizes = np.asarray([np.round((num_cells[i] / np.sum(num_cells)) * miniBatch_size).astype(int) for i in range(num_files)])

    # determine total number of iterations
    total_iters = np.floor(num_cells[0] * max_epochs / miniBatch_sizes[0]).astype(int)

    # generate all slicing index for each iteration
    all_idx = [_generate_idx(total_iters, miniBatch_sizes[i], max_epochs, h5_chunk_size, num_cells[i]) for i in range(num_files)]

    epoch = 0
    for iter in tqdm(range(total_iters)):
        # track epochs
        epoch_prev = epoch
        epoch = ((iter+1) * miniBatch_sizes[0]) // num_cells[0]

        # cells from each dataset that are used in this iteration
        Xs_miniBatch = []
        for i in range(num_files):
            miniBatch_idx = all_idx[i][iter]
            X_miniBatch = vstack([Xs[i]['scale_data'][left:right] for left, right in miniBatch_idx])
            #X_miniBatch = vstack([Xs[i][left:right] for left, right in miniBatch_idx]) ###########
            Xs_miniBatch.append(X_miniBatch.transpose().toarray())

        ## 1) update H_i_minibatch by ANLS
        H_miniBatch = []
        for i in range(num_files):
            H_miniBatch.append(nnlsm_blockpivot(A=np.vstack(((W + Vs[i]), np.sqrt(value_lambda) * Vs[i])),
                                                B=np.vstack((Xs_miniBatch[i], np.zeros((num_genes, miniBatch_sizes[i])))))[0])

        ## 2) updata A and B matrices
        for i in range(num_files):
            A[i], B[i], A_old[i], B_old[i] = _update_A_B(A[i], B[i], A_old[i], B_old[i], H_miniBatch[i], Xs_miniBatch[i], miniBatch_sizes[i], iter, epoch, epoch_prev)

        ## 3) update W, V_i by HALS
        iter_miniBatch = 1
        while iter_miniBatch <= miniBatch_max_iters:
            W = _update_W_HALS(A, B, W, Vs)  # update W
            Vs = _update_V_HALS(A, B, W, Vs, value_lambda)  # update V_i

            iter_miniBatch += 1

    # Calculate metagene loadings (H metrics)
    if verbose:
        print('Calculate metagene loadings...')

    Hs = []
    for i in range(num_files):
        num_batch = np.ceil(num_cells[i] / miniBatch_size).astype(int)
        H_miniBatch = []
        for batch in range(num_batch):
            if batch != num_batch:
                X_miniBatch = Xs[i]['scale_data'][batch*miniBatch_size:(batch+1)*miniBatch_size].transpose().toarray()
                #X_miniBatch = Xs[i][batch * miniBatch_size:(batch + 1) * miniBatch_size].transpose().toarray()
            else:
                X_miniBatch = Xs[i]['scale_data'][batch*miniBatch_size:num_cells[i]].transpose().toarray()
                #X_miniBatch = Xs[i][batch * miniBatch_size:num_cells[i]].transpose().toarray()

            H_miniBatch.append(nnlsm_blockpivot(A=np.vstack(((W + Vs[i]), np.sqrt(value_lambda) * Vs[i])),
                                                B=np.vstack((X_miniBatch, np.zeros((num_genes, X_miniBatch.shape[1])))))[0])

        Hs.append(np.hstack(H_miniBatch))
        
        
        ### 1. Initialization (W, V_i, A(HiHi^t), B(XiHit))
        np.random.seed(seed=rand_seed)
        W = _init_W(num_genes, k, rand_seed) if W_init is None else W_init
        V = _init_V_online(num_cells, k, X, h5_chunk_size, rand_seed) if V_init is None else V_init
        A = np.zeros((k, k)) if A_init is None else A_init  # A = HiHi^t
        B = np.zeros((num_genes, k)) if B_init is None else B_init  # B = XiHit

        # save information older than 2 epochs
        A_old = np.zeros((k, k))
        B_old = np.zeros((num_genes, k))

        # determine total number of iterations
        total_iters = np.floor(num_cells[0] * max_epochs / miniBatch_size).astype(int)

        # generate all slicing index for each iteration
        all_idx = _generate_idx(total_iters, miniBatch_size, max_epochs, h5_chunk_size, num_cells[i])

        ### 2. Iteration starts here
        epoch = 0
        for iter in tqdm(range(total_iters)):
            # track epochs
            epoch_prev = epoch
            epoch = ((iter+1) * miniBatch_size) // num_cells[i]

            # cells that are used in this iteration
            miniBatch_idx = all_idx[iter]
            X_miniBatch = vstack([X['scale_data'][left:right] for left, right in miniBatch_idx])
            # X_miniBatch = vstack([X[left:right] for left, right in miniBatch_idx]) ###########
            X_miniBatch = X_miniBatch.transpose().toarray()

            ## 1) update H_i_minibatch by ANLS
            H_miniBatch = nnlsm_blockpivot(A=np.vstack(((W + V), np.sqrt(value_lambda) * V)),
                                           B=np.vstack((X_miniBatch, np.zeros((num_genes, miniBatch_size)))))[0]

            ## 2) updata A and B matrices
            A, B, A_old, B_old = _update_A_B(A, B, A_old, B_old, H_miniBatch, X_miniBatch, miniBatch_size, iter, epoch, epoch_prev)

            ## 3) update W, V_i by HALS
            iter_miniBatch = 1
            while iter_miniBatch <= miniBatch_max_iters:
                W = _update_W_HALS([A], [B], W, [V])  # update W
                V = _update_V_HALS([A], [B], W, [V], value_lambda)[0]  # update V

                iter_miniBatch += 1


        # Calculate metagene loadings (H metrics)
        if verbose:
            print('Calculate metagene loadings...')

        Hs = []
        for i in range(num_files):
            num_batch = np.ceil(num_cells[i] / miniBatch_size).astype(int)
            H_miniBatch = []
            for batch in range(num_batch):
                if batch != num_batch:
                    X_miniBatch = Xs[i]['scale_data'][batch*miniBatch_size:(batch+1)*miniBatch_size].transpose().toarray()
                    #X_miniBatch = Xs[i][
                    #              batch * miniBatch_size:(batch + 1) * miniBatch_size].transpose().toarray()
                else:
                    X_miniBatch = Xs[i]['scale_data'][batch*miniBatch_size:num_cells[i]].transpose().toarray()
                    #X_miniBatch = Xs[i][batch * miniBatch_size:num_cells[i]].transpose().toarray()

                H_miniBatch.append(nnlsm_blockpivot(A=np.vstack(((W + Vs[i]), np.sqrt(value_lambda) * Vs[i])),
                                                    B=np.vstack((X_miniBatch, np.zeros((num_genes, X_miniBatch.shape[1])))))[0])

            Hs.append(np.hstack(H_miniBatch))
            
                    # Close all files in the end
        X.close()

        # Sava results and close hdf5 files
        X_new[i].obsm['H'] = Hs[i].transpose()
        X_new[i].varm['W'] = W
        X_new[i].varm['V'] = V

"""



import time
import numpy as np

from ._utilities import _init_W, _init_H, _init_V, _update_H_HALS, _update_W_HALS, _update_V_HALS

def iNMF_HALS(liger_object,
              k=20,
              value_lambda=5.0,
              thresh=1e-4,
              max_iters=25,
              nrep=1,
              rand_seed=1):
    """Perform iNMF on scaled datasets using HALS method

    Parameters
    ----------
    liger_object : liger object
        Should normalize, select genes, and scale before calling.
    k : int, optional
        Inner dimension of factorization (number of factors). Run suggestK to determine
        appropriate value; a general rule of thumb is that a higher k will be needed for datasets with
        more sub-structure (the default is 20).
    value_lambda : float, optional
        Regularization parameter. Larger values penalize dataset-specific effects more
        strongly (ie. alignment should increase as lambda increases). Run suggestLambda to determine
        most appropriate value for balancing dataset alignment and agreement (the default is 5.0).
    thresh : float, optional
        Convergence threshold. Convergence occurs when |obj0-obj|/(mean(obj0,obj)) < thresh
        (the default is 1e-4).
    max_iters : int, optional
        Maximum number of block coordinate descent iterations to perform (the default is 25).
    nrep : int, optional
        Number of restarts to perform (iNMF objective function is non-convex, so taking the
        best objective from multiple successive initializations is recommended). For easier
        reproducibility, this increments the random seed by 1 for each consecutive restart, so future
        factorizations of the same dataset can be run with one rep if necessary (the default is 1).
    rand_seed : int, optional
        Random seed to allow reproducible results (the default is 1).

    Returns
    -------
    liger_object : liger object
        liger object with H, W, and V annotations.

    Examples
    --------
    >>> adata1 = AnnData(np.arange(12).reshape((4, 3)))
    >>> adata2 = AnnData(np.arange(12).reshape((4, 3)))
    >>> ligerex = pyliger.create_liger([adata1, adata2])
    >>> pyliger.normalize(ligerex)
    >>> pyliger.select_genes(ligerex) # select genes
    >>> pyliger.scale_not_center(ligerex)
    >>> pyliger.iNMF_HALS(ligerex, k = 20, value_lambda = 5, nrep = 3)
    """

    num_samples = len(liger_object.adata_list)
    X = [adata.layers['scale_data'].transpose() for adata in liger_object.adata_list]

    num_cells = [X[i].shape[1] for i in range(num_samples)]
    num_genes = X[0].shape[0]

    # set seed
    np.random.seed(seed=rand_seed)

    # allow restarts
    num_rep = 0
    while num_rep < nrep:
        # Initialization
        V = _init_V(num_cells, num_samples, k, X)
        W = _init_W(num_genes, k, rand_seed)
        H = _init_H(num_cells, num_samples, k)

        # Initial training obj
        obj_train_approximation = 0
        obj_train_penalty = 0
        for i in range(num_samples):
            obj_train_approximation += np.linalg.norm(X[i].toarray() - (W + V[i]) @ H[i]) ** 2
            obj_train_penalty += np.linalg.norm(V[i] @ H[i]) ** 2

        obj_train = obj_train_approximation + value_lambda * obj_train_penalty

        # print start information
        print('Initial Training Obj: {}'.format(obj_train))

        iteration = 1
        total_time = 0  # track the total amount of time used for learning
        delta = np.inf

        # Perform block coordinate descent
        while delta > thresh and iteration <= max_iters:

            iter_start_time = time.time()

            # update H matrix
            H = _update_H_HALS(H, V, W, X, value_lambda)
            #VitVi = [V[i].transpose() @ V[i] for i in range(num_samples)]
            #W_Vi = [W + V[i] for i in range(num_samples)]
            #W_Vi_sq = [W_Vi[i].transpose() @ W_Vi[i] for i in range(num_samples)]
            #for i in range(num_samples):
            #    for j in range(k):
            #        H[i][j, :] = nonneg(H[i][j, :] + (
            #                    W_Vi[i][:, j].transpose() @ X[i] - W_Vi[i][:, j].transpose() @ W_Vi[i] @ H[
            #                i] - value_lambda * VitVi[i][j, :] @ H[i]) / (
            #                                        W_Vi_sq[i][j, j] + value_lambda * VitVi[i][j, j]))

            # update W matrix
            XHt = [X[i] @ H[i].transpose() for i in range(num_samples)]
            HHt = [H[i] @ H[i].transpose() for i in range(num_samples)]
            #A = List()
            #[A.append(x) for x in HHt]
            #B = List()
            #[B.append(x) for x in XHt]
            #temp_V = List()
            #[temp_V.append(x) for x in V]
            W = _update_W_HALS(HHt, XHt, W, V)

            # update V matrix
            V = _update_V_HALS(HHt, XHt, W, V, value_lambda)

            # training obj
            obj_train_prev = obj_train
            obj_train_approximation = 0
            obj_train_penalty = 0
            for i in range(num_samples):
                obj_train_approximation += np.linalg.norm(X[i] - (W + V[i]) @ H[i]) ** 2
                obj_train_penalty += np.linalg.norm(V[i] @ H[i]) ** 2
            obj_train = obj_train_approximation + value_lambda * obj_train_penalty
            delta = np.absolute(obj_train_prev - obj_train) / ((obj_train_prev + obj_train) / 2)

            # print result information
            total_time += (time.time() - iter_start_time)
            print('Iter: {}, Total time: {}, Training Obj: {}'.format(iteration, total_time, obj_train))

            if obj_train <= obj_train_prev:
                final_H = H
                final_W = W
                final_V = V

            iteration += 1

        num_rep += 1

    liger_object.W = final_W
    # Save results into the liger_object
    for i in range(num_samples):
        liger_object.adata_list[i].obsm['H'] = final_H[i].transpose()
        liger_object.adata_list[i].varm['W'] = final_W
        liger_object.adata_list[i].varm['V'] = final_V[i]

    return None



import os
import logging
import h5sparse
import numpy as np

from anndata import AnnData


def _merge_sparse_data_all(adata_list, library_names=None):
    """ Function to merge all sparse data into a single one

    Function takes in a list of DGEs, with gene row names and cell column names,
    and merges them into a single DGE.
    Also adds library_names to cell_names if expected to be overlap (common with 10X barcodes)

    Parameters
    ----------
    adata_list : AnnData
        List of AnnData objects which store expression matrices (gene by cell).
    library_names : list, optional
        (the default is None)

    Returns
    -------
    merged_adata : AnnData
        AnnData object stores expression matrix across datasets
    """

    merged_adata = AnnData()

    for adata in adata_list:
        merged_adata = merged_adata.concatenate(adata, join='inner')

    return merged_adata


def _remove_missing_obs(adata,
                        slot_use='raw_data',
                        use_rows=True):
    """Remove cells/genes with no expression across any genes/cells

    Removes cells/genes from chosen slot with no expression in any genes or cells respectively.

    Parameters
    ----------
    TODO: change to adata
    adata : AnnData object
        object (scale_data or norm_data must be set).
    slot_use : str, optional, 'raw_data' or 'scale_data'
        The data slot to filter (the default is 'raw_data').
    use_rows : bool, optional
        Treat each row as a cell (the default is True).

    Returns
    -------
    liger_object : liger object
        object with modified raw_data (or chosen slot) (dataset names preserved).

    Examples
    --------
    >>> adata = _remove_missing_obs(adata)
    """
    removed = str(np.where(slot_use in ['raw_data', 'norm_data'] and use_rows == True, 'cells', 'genes'))
    expressed = str(np.where(removed == 'cells', ' any genes', ''))

    data_type = adata.uns['sample_name']
    if slot_use == 'raw_data':
        filter_data = adata.X
    elif slot_use == 'scale_data':
        filter_data = adata.layers['scale_data']

    if use_rows:
        missing = np.ravel(np.sum(filter_data, axis=1)) == 0
    else:
        missing = np.ravel(np.sum(filter_data, axis=0)) == 0
    if np.sum(missing) > 0:
        #logging.info('Removing {} {} not expressing{} in {}.'.format(np.sum(missing), removed, expressed, data_type))
        print('Removing {} {} not expressing{} in {}.'.format(np.sum(missing), removed, expressed, data_type))
        if use_rows:
            # show gene name when the total of missing is less than 25
            if np.sum(missing) < 25:
                print(adata.obs.index[missing])
            adata = adata[~missing, :].copy()
        else:
            # show cell name when the total of missing is less than 25
            if np.sum(missing) < 25:
                print(adata.var.index[missing])
            adata = adata[:, ~missing].copy()

    return adata


################################## For Use of hdf5 ################################
def _h5_idx_generator(chunk_size, matrix_size):
    """ """
    previous_idx = 0
    if matrix_size < chunk_size:
        current_idx = matrix_size
    else:
        current_idx = chunk_size
    num_chunk = np.ceil(matrix_size/chunk_size).astype(int)
    iter = 0
    while current_idx <= matrix_size and iter < num_chunk:
        yield int(previous_idx), int(current_idx)
        previous_idx += chunk_size
        current_idx += chunk_size
        if current_idx > matrix_size:
            current_idx = matrix_size
        iter += 1
    return None

def merge_H5(file_list,
             library_names,
             new_filename,
             format_type='10X',
             data_name=None,
             genes_name=None,
             barcodes_name=None):
    return None


def nonneg(x, eps=1e-16):
    """ Given a input matrix, set all negative values to be zero """
    x[x < eps] = eps
    return x


def _create_h5_using_adata(adata, chunk_size):
    # create h5 file.
    if not os.path.isdir('./results'):
        os.mkdir('./results')

    file_name = './results/' + adata.uns['sample_name'] + '.hdf5'
    with h5sparse.File(file_name, 'w') as f:
        for left, right in _h5_idx_generator(chunk_size, adata.shape[0]):
            if 'raw_data' not in f.keys():
                f.create_dataset('raw_data', data=adata[left:right, :].X, chunks=(chunk_size,), maxshape=(None,))
            else:
                f['raw_data'].append(adata[left:right, :].X)
    return None


import numpy as np
import pandas as pd

from pathlib import Path
from anndata import AnnData
from collections import Counter
from scipy.sparse import csr_matrix


def make_feature_matrix(file_dir, gene_file, promoter_file, filter_threshold, return_separate):
    """

    :param file_dir:
    :param gene_file:
    :param promoter_file:
    :param filter_threshold:
    :return:
    """
    ### 0. build file path
    p_gene = Path(file_dir, gene_file)
    p_promoter = Path(file_dir, promoter_file)

    ### 1. extract barcodes (equivalent to step 3 from R tutorial)
    genes_bc = pd.read_csv(p_gene, sep='\t', comment='t', header=None, usecols=[3, 6])
    promoters_bc = pd.read_csv(p_promoter, sep='\t', comment='t', header=None, usecols=[3, 6])
    genes_bc = genes_bc.sort_values(by=[3])
    promoters_bc = promoters_bc.sort_values(by=[3])
    num_genes = genes_bc.shape[0]

    ### 2. make matrices for gene counts and promoter counts (equivalent to step 4 from R tutorial)
    genes_bc_total, genes_bc_counts, genes_indptr, _ = _extract_barcodes(genes_bc)
    promoters_bc_total, promoters_bc_counts, _, promoters_num_bc = _extract_barcodes(promoters_bc)

    genes_bc_unique, genes_bc_inverse = np.unique(genes_bc_total, return_inverse=True)

    # select barcodes
    genes_mtx = csr_matrix((genes_bc_counts, genes_bc_inverse, genes_indptr), shape=(num_genes, len(genes_bc_unique)))
    filter_idx = np.sum(genes_mtx, axis=0) > filter_threshold
    genes_mtx = genes_mtx[:, filter_idx.A1]

    row_idx = np.repeat(np.arange(num_genes), promoters_num_bc)
    mask = np.isin(promoters_bc_total, genes_bc_unique[filter_idx.A1])

    promoters_bc_total = promoters_bc_total[mask]
    promoters_bc_counts = promoters_bc_counts[mask]
    row_idx = row_idx[mask]
    promoters_bc_unique, promoters_bc_inverse = np.unique(promoters_bc_total, return_inverse=True)
    promoters_mtx = csr_matrix((promoters_bc_counts, (row_idx, promoters_bc_inverse)), shape=(num_genes, len(promoters_bc_unique)))

    if return_separate:
        return None
    else:
        sum_mtx = genes_mtx + promoters_mtx
        barcodes = pd.DataFrame(genes_bc_unique[filter_idx.A1])
        barcodes.columns = ['barcodes']
        barcodes = barcodes.set_index('barcodes')

        gene_name = pd.DataFrame(genes_bc[3])
        gene_name.columns = ['gene_name']
        gene_name = gene_name.set_index('gene_name')

        atac_anndata = AnnData(csr_matrix(sum_mtx.transpose()), obs=barcodes, var=gene_name,
                uns={'sample_name': 'test_1', 'data_type': 'ATAC'})
        return atac_anndata


def _extract_barcodes(bedmat):
    """"""
    barcodes_total = []
    barcodes_value = []
    indptr = [0]
    num_barcodes = []
    for i in range(bedmat.shape[0]):
        row = bedmat.iloc[i]

        if type(row[6]) == str:
            barcodes = row[6].split(';')
        else:
            barcodes = []

        barcodes_count = Counter(barcodes)
        barcodes_total.extend(list(barcodes_count))
        barcodes_value.extend(list(barcodes_count.values()))
        indptr.append(indptr[i]+len(barcodes_count))
        num_barcodes.append(len(barcodes_count))

    return np.asarray(barcodes_total), np.asarray(barcodes_value), indptr, num_barcodes




import numpy as np
import scipy.sparse as sps
import numpy.linalg as nla
import scipy.linalg as sla
from numba import jit, njit, prange
from .._utilities import _h5_idx_generator


def nonneg(x, eps=1e-16):
    """ Given a input matrix, set all negative values to be zero """
    x[x < eps] = eps
    return x


def _init_W(num_genes, k, rand_seed):
    """helper function to initialize a W matrix"""
    # set seed
    np.random.seed(seed=rand_seed)

    W = np.abs(np.random.uniform(0, 2, (num_genes, k)))

    # normalize columns of dictionaries
    W = W / np.sqrt(np.sum(np.square(W), axis=0))

    return W


def _init_V(num_cells, num_samples, k, Xs):
    """helper function to initialize a V matrix for in-memory mode"""
    # pick k sample from datasets as initial V matrix
    V = [Xs[i][:, np.random.choice(list(range(num_cells[i])), k)].toarray() for i in range(num_samples)]

    # normalize columns of dictionaries
    V = [V[i] / np.sqrt(np.sum(np.square(V[i]), axis=0)) for i in range(num_samples)]

    return V
"""
def _init_V_online(num_cells, num_samples, k, Xs, chunk_size, rand_seed):

    np.random.seed(seed=rand_seed)

    Vs = []
    for i in range(num_samples):
        # pick k sample from datasets as initial H matrix
        idx = np.sort(np.random.choice(list(range(num_cells[i])), k))
        V = []
        for left, right in _h5_idx_generator(chunk_size, num_cells[i]):
            select_idx = idx[(idx >= left) & (idx < right)] - left  # shift index because of handling chunk each time
            if select_idx.shape[0] > 0:  # only load chunks whose indexes are picked
                X = Xs[i]['scale_data'][left:right]
                V.append(X[select_idx, :])
        V = sps.vstack(V).transpose().toarray()

        # normalize columns of dictionaries
        V = V / np.sqrt(np.sum(np.square(V), axis=0))

        Vs.append(V)

    return Vs
"""

def _init_V_online(num_cell, k, X, chunk_size, rand_seed):
    """helper function to initialize a V matrix for online learning"""
    np.random.seed(seed=rand_seed)

    # pick k sample from datasets as initial H matrix
    idx = np.sort(np.random.choice(list(range(num_cell)), k))
    V = []
    for left, right in _h5_idx_generator(chunk_size, num_cell):
        select_idx = idx[(idx >= left) & (idx < right)] - left  # shift index because of handling chunk each time
        if select_idx.shape[0] > 0:  # only load chunks whose indexes are picked
            X_chunk = X['scale_data'][left:right]
            V.append(X_chunk[select_idx, :])
    V = sps.vstack(V).transpose().toarray()

    # normalize columns of dictionaries
    V = V / np.sqrt(np.sum(np.square(V), axis=0))

    return V


def _init_H(num_cells, num_samples, k):
    """helper function to initialize a H matrix"""
    H = [np.random.uniform(0, 2, (k, num_cells[i])) for i in range(num_samples)]
    return H

def _init_Hi():
    """"""
    return None

def _update_W_HALS(A, B, W, V):
    """helper function to update W matrix by HALS
    A = HiHi^t, B = XiHit, W = gene x k, V = [gene x k]"""
    for j in range(W.shape[1]):
        W_update_numerator = np.zeros(W.shape[0])
        W_update_denominator = 0.0
        for i in range(len(V)):
            W_update_numerator = W_update_numerator + B[i][:, j] - ((W + V[i]) @ A[i])[:, j]
            W_update_denominator += A[i][j, j]
        W[:, j] = nonneg(W[:, j] + W_update_numerator / W_update_denominator)

    return W


def _update_V_HALS(A, B, W, V, value_lambda):
    """helper function to update V matrix by HALS
    A = HiHi^t, B = XiHit, W = gene x k, V = [gene x k]"""
    for j in range(W.shape[1]):
        for i in range(len(V)):
            V[i][:, j] = nonneg(V[i][:, j] + (B[i][:, j] - (W + (1 + value_lambda) * V[i]) @ A[i][:, j]) / ((1 + value_lambda) * A[i][j, j]))

    return V


def _update_H_HALS(H, V, W, X, value_lambda):
    """helper function to update H matrix by HALS"""
    VitVi = [Vi.transpose() @ Vi for Vi in V]
    W_Vi = [W + Vi for Vi in V]
    W_Vi_sq = [W_Vii.transpose() @ W_Vii for W_Vii in W_Vi]
    for i in range(len(V)):
        for j in range(W.shape[1]):
            H[i][j, :] = nonneg(H[i][j, :] + (
                    W_Vi[i][:, j].transpose() @ X[i] - W_Vi[i][:, j].transpose() @ W_Vi[i] @ H[
                i] - value_lambda * VitVi[i][j, :] @ H[i]) / (
                                        W_Vi_sq[i][j, j] + value_lambda * VitVi[i][j, j]))
    return H


######use numba
"""
#@jit(nopython=True)
def _update_W_HALS(A, B, W, V):
    
    for j in range(W.shape[1]):
        W_update_numerator = np.zeros(W.shape[0])
        W_update_denominator = 0.0
        for i in range(len(V)):
            W_update_numerator = W_update_numerator + B[i][:, j] - ((W + V[i]) @ A[i])[:, j]
            W_update_denominator += A[i][j, j]
        temp = W[:, j] + W_update_numerator / W_update_denominator
        temp[temp < 1e-16] = 1e-16
        W[:, j] = temp
    return W


#@jit(nopython=True)
def _update_V_HALS(A, B, W, V, value_lambda):

    for j in range(V[0].shape[1]):
        for i in range(len(V)):
            temp = V[i][:, j] + (B[i][:, j] - (W + (1 + value_lambda) * V[i]) @ A[i][:, j]) / ((1 + value_lambda) * A[i][j, j])
            temp[temp < 1e-16] = 1e-16
            V[i][:, j] = temp

    return V
"""
def nnlsm_blockpivot(A, B, is_input_prod=False, init=None):
    """ Nonnegativity-constrained least squares with block principal pivoting method and column grouping
    Solves min ||AX-B||_2^2 s.t. X >= 0 element-wise.
    J. Kim and H. Park, Fast nonnegative matrix factorization: An active-set-like method and comparisons,
    SIAM Journal on Scientific Computing,
    vol. 33, no. 6, pp. 3261-3281, 2011.
    Parameters
    ----------
    A : numpy.array, shape (m,n)
    B : numpy.array or scipy.sparse matrix, shape (m,k)
    Optional Parameters
    -------------------
    is_input_prod : True/False. -  If True, the A and B arguments are interpreted as
            AtA and AtB, respectively. Default is False.
    init: numpy.array, shape (n,k). - If provided, init is used as an initial value for the algorithm.
            Default is None.
    Returns
    -------
    X, (success, Y, num_cholesky, num_eq, num_backup)
    X : numpy.array, shape (n,k) - solution
    success : True/False - True if the solution is found. False if the algorithm did not terminate
            due to numerical errors.
    Y : numpy.array, shape (n,k) - Y = A.T * A * X - A.T * B
    num_cholesky : int - the number of Cholesky factorizations needed
    num_eq : int - the number of linear systems of equations needed to be solved
    num_backup: int - the number of appearances of the back-up rule. See SISC paper for details.
    """
    if is_input_prod:
        AtA = A
        AtB = B
    else:
        AtA = A.T.dot(A)
        if sps.issparse(B):
            AtB = B.T.dot(A)
            AtB = AtB.T
        else:
            AtB = A.T.dot(B)

    (n, k) = AtB.shape
    MAX_ITER = n * 5

    if init is not None:
        PassSet = init > 0
        X, num_cholesky, num_eq = normal_eq_comb(AtA, AtB, PassSet)
        Y = AtA.dot(X) - AtB
    else:
        X = np.zeros([n, k])
        Y = -AtB
        PassSet = np.zeros([n, k], dtype=bool)
        num_cholesky = 0
        num_eq = 0

    p_bar = 3
    p_vec = np.zeros([k])
    p_vec[:] = p_bar
    ninf_vec = np.zeros([k])
    ninf_vec[:] = n + 1
    not_opt_set = np.logical_and(Y < 0, ~PassSet)
    infea_set = np.logical_and(X < 0, PassSet)

    not_good = np.sum(not_opt_set, axis=0) + np.sum(infea_set, axis=0)
    not_opt_colset = not_good > 0
    not_opt_cols = not_opt_colset.nonzero()[0]

    big_iter = 0
    num_backup = 0
    success = True
    while not_opt_cols.size > 0:
        big_iter += 1
        if MAX_ITER > 0 and big_iter > MAX_ITER:
            success = False
            break

        cols_set1 = np.logical_and(not_opt_colset, not_good < ninf_vec)
        temp1 = np.logical_and(not_opt_colset, not_good >= ninf_vec)
        temp2 = p_vec >= 1
        cols_set2 = np.logical_and(temp1, temp2)
        cols_set3 = np.logical_and(temp1, ~temp2)

        cols1 = cols_set1.nonzero()[0]
        cols2 = cols_set2.nonzero()[0]
        cols3 = cols_set3.nonzero()[0]

        if cols1.size > 0:
            p_vec[cols1] = p_bar
            ninf_vec[cols1] = not_good[cols1]
            true_set = np.logical_and(not_opt_set, np.tile(cols_set1, (n, 1)))
            false_set = np.logical_and(infea_set, np.tile(cols_set1, (n, 1)))
            PassSet[true_set] = True
            PassSet[false_set] = False
        if cols2.size > 0:
            p_vec[cols2] = p_vec[cols2] - 1
            temp_tile = np.tile(cols_set2, (n, 1))
            true_set = np.logical_and(not_opt_set, temp_tile)
            false_set = np.logical_and(infea_set, temp_tile)
            PassSet[true_set] = True
            PassSet[false_set] = False
        if cols3.size > 0:
            for col in cols3:
                candi_set = np.logical_or(
                    not_opt_set[:, col], infea_set[:, col])
                to_change = np.max(candi_set.nonzero()[0])
                PassSet[to_change, col] = ~PassSet[to_change, col]
                num_backup += 1

        (X[:, not_opt_cols], temp_cholesky, temp_eq) = normal_eq_comb(
            AtA, AtB[:, not_opt_cols], PassSet[:, not_opt_cols])
        num_cholesky += temp_cholesky
        num_eq += temp_eq
        X[abs(X) < 1e-16] = 0
        Y[:, not_opt_cols] = AtA.dot(X[:, not_opt_cols]) - AtB[:, not_opt_cols]
        Y[abs(Y) < 1e-16] = 0

        not_opt_mask = np.tile(not_opt_colset, (n, 1))
        not_opt_set = np.logical_and(
            np.logical_and(not_opt_mask, Y < 0), ~PassSet)
        infea_set = np.logical_and(
            np.logical_and(not_opt_mask, X < 0), PassSet)
        not_good = np.sum(not_opt_set, axis=0) + np.sum(infea_set, axis=0)
        not_opt_colset = not_good > 0
        not_opt_cols = not_opt_colset.nonzero()[0]

    return X, (success, Y, num_cholesky, num_eq, num_backup)


def normal_eq_comb(AtA, AtB, PassSet=None):
    """ Solve many systems of linear equations using combinatorial grouping.
    M. H. Van Benthem and M. R. Keenan, J. Chemometrics 2004; 18: 441-450
    Parameters
    ----------
    AtA : numpy.array, shape (n,n)
    AtB : numpy.array, shape (n,k)
    Returns
    -------
    (Z,num_cholesky,num_eq)
    Z : numpy.array, shape (n,k) - solution
    num_cholesky : int - the number of unique cholesky decompositions done
    num_eq: int - the number of systems of linear equations solved
    """
    num_cholesky = 0
    num_eq = 0
    if AtB.size == 0:
        Z = np.zeros([])
    elif (PassSet is None) or np.all(PassSet):
        Z = nla.solve(AtA, AtB)
        num_cholesky = 1
        num_eq = AtB.shape[1]
    else:
        Z = np.zeros(AtB.shape)
        if PassSet.shape[1] == 1:
            if np.any(PassSet):
                cols = PassSet.nonzero()[0]
                Z[cols] = nla.solve(AtA[np.ix_(cols, cols)], AtB[cols])
                num_cholesky = 1
                num_eq = 1
        else:
            #
            # Both _column_group_loop() and _column_group_recursive() work well.
            # Based on preliminary testing,
            # _column_group_loop() is slightly faster for tiny k(<10), but
            # _column_group_recursive() is faster for large k's.
            #
            grps = _column_group_recursive(PassSet)
            for gr in grps:
                cols = PassSet[:, gr[0]].nonzero()[0]
                if cols.size > 0:
                    ix1 = np.ix_(cols, gr)
                    ix2 = np.ix_(cols, cols)
                    #
                    # scipy.linalg.cho_solve can be used instead of numpy.linalg.solve.
                    # For small n(<200), numpy.linalg.solve appears faster, whereas
                    # for large n(>500), scipy.linalg.cho_solve appears faster.
                    # Usage example of scipy.linalg.cho_solve:
                    #Z[ix1] = sla.cho_solve(sla.cho_factor(AtA[ix2]),AtB[ix1])
                    #
                    Z[ix1] = nla.solve(AtA[ix2], AtB[ix1])
                    num_cholesky += 1
                    num_eq += len(gr)
                    num_eq += len(gr)
    return Z, num_cholesky, num_eq


def _column_group_recursive(B):
    """ Given a binary matrix, find groups of the same columns
        with a recursive strategy
    Parameters
    ----------
    B : numpy.array, True/False in each element
    Returns
    -------
    A list of arrays - each array contain indices of columns that are the same.
    """
    initial = np.arange(0, B.shape[1])
    return [a for a in column_group_sub(B, 0, initial) if len(a) > 0]


def column_group_sub(B, i, cols):
    vec = B[i][cols]
    if len(cols) <= 1:
        return [cols]
    if i == (B.shape[0] - 1):
        col_trues = cols[vec.nonzero()[0]]
        col_falses = cols[(~vec).nonzero()[0]]
        return [col_trues, col_falses]
    else:
        col_trues = cols[vec.nonzero()[0]]
        col_falses = cols[(~vec).nonzero()[0]]
        after = column_group_sub(B, i + 1, col_trues)
        after.extend(column_group_sub(B, i + 1, col_falses))
    return after



import numpy as np
import igraph as ig
from annoy import AnnoyIndex
from pynndescent import NNDescent
from scipy.sparse import csr_matrix
from sklearn.neighbors import NearestNeighbors


def run_knn(H, k):
    """ """
    neigh = NearestNeighbors(n_neighbors=k, radius=0, algorithm='kd_tree')
    neigh.fit(H)
    H_knn = neigh.kneighbors(H, n_neighbors=k, return_distance=False)

    return H_knn


def run_ann(H, k, num_trees=None):

    # implementation using annoy library
    num_observations = H.shape[0]

    # decide number of trees
    if num_trees is None:
        if num_observations < 100000:
            num_trees = 10
        elif num_observations < 1000000:
            num_trees = 20
        elif num_observations < 5000000:
            num_trees = 50
        else:
            num_trees = 100

    # build knn graph
    t = AnnoyIndex(k, 'angular')
    for i in range(num_observations):
        t.add_item(i, H[i])
    t.build(num_trees)

    # create knn indices matrices
    H_knn = np.vstack([t.get_nns_by_vector(H[i], k) for i in range(num_observations)])
    #for i in range(num_observations):
    #    if i == 0:
    #        H_knn = np.array(t.get_nns_by_vector(H[i], k))
    #    else:
    #        H_knn = np.vstack((H_knn, t.get_nns_by_vector(H[i], k)))

    return H_knn


def run_pynndescent(H, k, num_trees=None):

    num_observations = H.shape[0]

    index = NNDescent(H)
    H_knn = index.query(H, k=k)

    return H_knn


@njit
def cluster_vote(clusts, H_knn, k):
    """"""
    for i in range(H_knn.shape[0]):
        clust_counts = {}
        for j in range(k):
            if clusts[H_knn[i, j]] not in clust_counts:
                clust_counts[clusts[H_knn[i, j]]] = 1
            else:
                clust_counts[clusts[H_knn[i, j]]] += 1

        max_clust = -1
        max_count = 0
        for key, value in clust_counts.items():
            if value > max_count:
                max_clust = key
                max_count = value
            elif value == max_count:
                if key > max_clust:
                    max_clust = key
                    max_count = value
        clusts[i] = max_clust
    return clusts


def refine_clusts(H, clusts, k, use_ann, num_trees=None):
    """helper function for refining clusters related to function quantile_norm """
    if use_ann:
        H_knn = run_ann(H, k, num_trees)
    else:
        H_knn = run_knn(H, k)

    clusts = cluster_vote(clusts, H_knn, k)

    return clusts


#@njit
def compute_snn(knn, prune):
    """helper function to compute the SNN graph"""
    # int for indexing
    knn = knn.astype(np.int)

    k = knn.shape[1]
    num_cells = knn.shape[0]

    rows = np.repeat(list(range(num_cells)), k)
    columns = knn.flatten()
    data = np.repeat(1, num_cells * k)
    snn = csr_matrix((data, (rows, columns)), shape=(num_cells, num_cells))

    snn = snn @ snn.transpose()

    rows, columns = snn.nonzero()
    data = snn.data / (k + (k - snn.data))
    data[data < prune] = 0

    return csr_matrix((data, (rows, columns)), shape=(num_cells, num_cells))


def build_igraph(snn):
    """"""
    sources, targets = snn.nonzero()
    weights = snn[sources, targets]

    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph()
    g.add_vertices(snn.shape[0])
    g.add_edges(list(zip(sources, targets)))
    g.es['weight'] = weights

    return g


def _assign_cluster(cluster, liger_object):
    """helper function to assign cluster to each dataset after community detection"""
    cluster = np.ravel(cluster)
    idx = 0
    for i, adata in enumerate(liger_object.adata_list):
        liger_object.adata_list[i].obs['cluster'] = cluster[idx:(idx + adata.shape[0])]
        idx += adata.shape[0]
    return None



import numpy as np
import pandas as pd
from scipy.sparse import vstack
from sklearn.preprocessing import scale

from ..tools import calc_dataset_specificity, _wilcoxon

# Find shared and dataset-specific markers
def get_factor_markers(liger_object,
                       dataset1 = 0,
                       dataset2 = 1,
                       factor_share_thresh = 10,
                       dataset_specificity = None,
                       log_fc_thresh = 1,
                       umi_thresh = 30,
                       frac_thresh = 0,
                       pval_thresh = 0.05,
                       num_genes = 30,
                       print_genes = False):
    """Find shared and dataset-specific markers

    Applies various filters to genes on the shared (W) and dataset-specific (V)  of the
    factorization, before selecting those which load most significantly on each factor (in a shared
    or dataset-specific way).

    Parameters
    ----------
    liger_object : liger
        Should call optimizeALS before calling.
    dataset1 : str
        Name of first dataset (default first dataset by order)
    dataset2 : str
        Name of second dataset (default second dataset by order)
    factor_share_thresh : list
        Use only factors with a dataset specificity less than or equal to threshold (default 10).
    dataset_specificity : TYPE
        Pre-calculated dataset specificity if available. Will calculate if not available.
    log_fc_thresh : float
        Lower log-fold change threshold for differential expression in markers (default 1).
    umi_thresh : int
        Lower UMI threshold for markers (default 30).
    frac_thresh : float
        Lower threshold for fraction of cells expressing marker (default 0).
    pval_thresh : float
        Upper p-value threshold for Wilcoxon rank test for gene expression (default 0.05).
    num_genes : int
        Max number of genes to report for each dataset (default 30).
    print_genes : boolean
        Print ordered markers passing logfc, umi and frac thresholds (default FALSE).
    """
    ### Parameter setting
    if isinstance(dataset1, str):
        dataset1 = liger_object.find_dataset_idx(dataset1)
    if isinstance(dataset2, str):
        dataset2 = liger_object.find_dataset_idx(dataset2)

    if num_genes is None:
        num_genes = len(liger_object.var_genes)

    if dataset_specificity is None:
        dataset_specificity = calc_dataset_specificity(liger_object, dataset1=dataset1, dataset2=dataset2, do_plot=False)

    factors_use = np.abs(dataset_specificity[2]) <= factor_share_thresh

    if len(factors_use) < 2:
        print('Warning: only {} factors passed the dataset specificity threshold.'.format(len(factors_use)))

    ### Extract values
    # scale H matrices
    Hs_scaled = [scale(matrix) for matrix in liger_object.H]

    # select index of max factor for each cell within the range of factors use
    labels = [np.argmax(H_scaled[:, factors_use], axis=1) for H_scaled in Hs_scaled]

    V1_matrices = []
    V2_matrices = []
    W_matrices = []
    for j in range(len(factors_use)):
        factor = factors_use[j]
        W = liger_object.W
        V1 = liger_object.V[dataset1]
        V2 = liger_object.V[dataset2]

        # if not max factor for any cell in either dataset
        if np.sum(labels[dataset1] == factor) <= 1 or np.sum(labels[dataset2] == factor) <= 1:
            print('Warning: factor {} did not appear as max in any cell in either dataset.'.format(factor))
            continue

        # Filter genes by gene_count and cell_frac thresholds
        expr_mat = []
        for dset in [dataset1, dataset2]:
            adata = liger_object.adata_list[dset]
            idx = labels[dset] == factor
            expr_mat.append(adata.layers['norm_data'][idx, :])
        expr_mat = vstack(expr_mat)
        cell_label = np.concatenate(np.repeat(dataset1, np.sum(labels[dataset1] == factor)), np.repeat(dataset1, np.sum(labels[dataset2] == factor)))
        wilcoxon_result = _wilcoxon(np.log(expr_mat.toarray() + 1e-10), cell_label)

        # filter based on log-fold change
        #log2fc = wilcoxon_result[wilcoxon_result['group'] == dataset1]['logFC'].to_numpy()
        filtered_genes_V1 = wilcoxon_result[(wilcoxon_result['logFC'] > log_fc_thresh) & (wilcoxon_result['pval'] < pval_thresh)]['feature'].to_numpy()
        filtered_genes_V2 = wilcoxon_result[(-wilcoxon_result['logFC'] > log_fc_thresh) & (wilcoxon_result['pval'] < pval_thresh)]['feature'].to_numpy()

        W = np.minimum((W+V1), (W+V2))
        V1 = V1[filtered_genes_V1, :]
        V2 = V2[filtered_genes_V2, :]

        if np.sum(filtered_genes_V1) == 0:
            top_genes_V1 = None
        else:
            top_genes_V1 = liger_object.adata_list[dataset1].var.index[np.argsort(V1[:, factor])[0:num_genes]]

        if len(filtered_genes_V2) == 0:
            top_genes_V2 = None
        else:
            top_genes_V2 = liger_object.adata_list[dataset2].var.index[np.argsort(V2[:, factor])[0:num_genes]]

        top_genes_W = liger_object.adata_list[dataset1].var.index[np.argsort(W[:, factor])[0:num_genes]]

        if print_genes:
            print('Factor: ', factor)
            print('Dataset 1')
            print(top_genes_V1)
            print('Shared')
            print(top_genes_W)
            print('Dataset 2')
            print(top_genes_V2)

        pvals = [] # order is V1, V2, W
        top_genes = [top_genes_V1, top_genes_V2, top_genes_W]

        for idx, gene_list in enumerate(top_genes):
            pvals.append(wilcoxon_result[(wilcoxon_result['feature'].isin(gene_list)) & (wilcoxon_result['group']==dataset1)]['pval'].to_numpy())

        V1_matrices.append(pd.DataFrame({'factor_num': np.repeat(factor, len(top_genes_V1)),
                                         'gene': top_genes_V1,
                                         'log2fc': wilcoxon_result[(wilcoxon_result['group'] == dataset1) & (wilcoxon_result['feature'].isin(top_genes_V1))]['logFC'].to_numpy(),
                                         'p_value': pvals[0]}))
        V2_matrices.append(pd.DataFrame({'factor_num': np.repeat(factor, len(top_genes_V2)),
                                         'gene': top_genes_V1,
                                         'log2fc': wilcoxon_result[(wilcoxon_result['group'] == dataset1) & (wilcoxon_result['feature'].isin(top_genes_V2))]['logFC'].to_numpy(),
                                         'p_value': pvals[1]}))
        W_matrices.append(pd.DataFrame({'factor_num': np.repeat(factor, len(top_genes_W)),
                                         'gene': top_genes_V1,
                                         'log2fc': wilcoxon_result[(wilcoxon_result['group'] == dataset1) & (wilcoxon_result['feature'].isin(top_genes_W))]['logFC'].to_numpy(),
                                         'p_value': pvals[2]}))

    V1_genes = pd.concat(V1_matrices)
    V2_genes = pd.concat(V2_matrices)
    W_genes = pd.concat(W_matrices)

    output_list = [V1_genes, W_genes, V2_genes]
    # cutoff only applies to dataset-specific dfs
    for idx, df in enumerate(output_list):
        if idx != 1:
            output_list[idx] = df[df['p_value'] < pval_thresh]

    return output_list

"""
def get_factor_markers(liger_object,
                       dataset1 = 0,
                       dataset2 = 1,
                       factor_share_thresh = 10,
                       dataset_specificity = None,
                       log_fc_thresh = 1,
                       umi_thresh = 30,
                       frac_thresh = 0,
                       pval_thresh = 0.05,
                       num_genes = 30,
                       print_genes = False):
    ### Parameter setting
    if isinstance(dataset1, str):
        dataset1 = liger_object.find_dataset_idx(dataset1)
    if isinstance(dataset2, str):
        dataset2 = liger_object.find_dataset_idx(dataset2)

    if num_genes is None:
        num_genes = len(liger_object.var_genes)

    if dataset_specificity is None:
        dataset_specificity = calc_dataset_specificity(liger_object, dataset1=dataset1, dataset2=dataset2, do_plot=False)

    factors_use = np.abs(dataset_specificity[2]) <= factor_share_thresh

    if len(factors_use) < 2:
        print('Warning: only {} factors passed the dataset specificity threshold.'.format(len(factors_use)))

    ### Extract values
    # scale H matrices
    Hs_scaled = [scale(matrix) for matrix in liger_object.H]

    # select index of max factor for each cell within the range of factors use
    labels = [np.argmax(H_scaled[:, factors_use], axis=1) for H_scaled in Hs_scaled]

    V1_matrices = []
    V2_matrices = []
    W_matrices = []
    for j in range(len(factors_use)):
        factor = factors_use[j]
        W = liger_object.W
        V1 = liger_object.V[dataset1]
        V2 = liger_object.V[dataset2]

        # if not max factor for any cell in either dataset
        if np.sum(labels[dataset1] == factor) <= 1 or np.sum(labels[dataset2] == factor) <= 1:
            print('Warning: factor {} did not appear as max in any cell in either dataset.'.format(factor))
            continue

        # Filter genes by gene_count and cell_frac thresholds
        gene_info = []
        for dset in [dataset1, dataset2]:
            adata = liger_object.adata_list[dset]
            idx = labels[dset] == factor
            gene_counts = np.ravel(np.sum(adata.X[idx, adata.uns['var_gene_idx']], axis=0))
            cell_fracs = np.ravel(np.sum(adata.X[idx, adata.uns['var_gene_idx']] > 0, axis=0)) / np.sum(idx)
            norm_counts = adata.layers['norm_data'][idx, :]
            mean_value = np.mean(norm_counts[:, adata.uns['var_gene_idx']], axis=0)
            gene_info.append(pd.DataFrame({'gene_counts': gene_counts, 'cell_fracs': cell_fracs, 'norm_counts': norm_counts, 'mean_value': mean_value}))

        # filter based on umi_thresh
        idx_gene_counts = np.logical_or(gene_info[0]['gene_counts'] > umi_thresh, gene_info[1]['gene_counts'] > umi_thresh)
        # filter based on fraction of cells
        idx_cell_fracs = np.logical_or(gene_info[0]['cell_fracs'] > frac_thresh, gene_info[1]['cell_fracs'] > frac_thresh)
        idx = np.logical_and(idx_gene_counts, idx_cell_fracs)  # bool index
        initial_filtered = liger_object.var_genes[idx]  # gene string name

        # filter based on log-fold change
        log2fc = np.log2(gene_info[0]['mean_value'] / gene_info[1]['mean_value'])
        idx_log2fc_v1 = log2fc > log_fc_thresh
        idx_log2fc_v2 = log2fc * (-1) > log_fc_thresh

        filtered_genes_V1 = np.logical_and(idx_log2fc_v1, idx) # bool index
        filtered_genes_V2 = np.logical_and(idx_log2fc_v2, idx) # bool index

        W = np.minimum((W+V1), (W+V2))
        V1 = V1[filtered_genes_V1, :]
        V2 = V2[filtered_genes_V2, :]

        if np.sum(filtered_genes_V1) == 0:
            top_genes_V1 = None
        else:
            top_genes_V1 = initial_filtered[np.argsort(V1[:, factor])[0:num_genes]] # gene name string

        if len(filtered_genes_V2) == 0:
            top_genes_V2 = None
        else:
            top_genes_V2 = initial_filtered[np.argsort(V2[:, factor])[0:num_genes]] # gene name string

        top_genes_W = initial_filtered[np.argsort(W[:, factor])[0:num_genes]]

        if print_genes:
            print('Factor: ', factor)
            print('Dataset 1')
            print(top_genes_V1)
            print('Shared')
            print(top_genes_W)
            print('Dataset 2')
            print(top_genes_V2)

        pvals = [] # order is V1, V2, W
        top_genes = [top_genes_V1, top_genes_V2, top_genes_W]

        for idx, gene_list in enumerate(top_genes):
            if gene_list is None:
                pvals.append(None)
                continue

            for gene in gene_list:
                idx_1 = liger_object.adata_list[dataset1].var.index.get_loc(gene)
                idx_2 = liger_object.adata_list[dataset2].var.index.get_loc(gene)
                pvals.append(mannwhitneyu(gene_info[dataset1]['norm_counts'][idx_1, :], gene_info[dataset2]['norm_counts'][idx_2, :]).pvalue)

        V1_matrices.append(pd.DataFrame([np.repeat(factor, len(top_genes_V1)),
                                         top_genes_V1,
                                         gene_info[dataset1]['gene_counts'][filtered_genes_V1],
                                         gene_info[dataset2]['gene_counts'][filtered_genes_V1],
                                         gene_info[dataset1]['cell_fracs'][filtered_genes_V1],
                                         gene_info[dataset2]['cell_fracs'][filtered_genes_V1],
                                         log2fc[filtered_genes_V1],
                                         pvals[0]]))
        V2_matrices.append(pd.DataFrame([np.repeat(factor, len(top_genes_V2)),
                                         top_genes_V1,
                                         gene_info[dataset1]['gene_counts'][filtered_genes_V2],
                                         gene_info[dataset2]['gene_counts'][filtered_genes_V2],
                                         gene_info[dataset1]['cell_fracs'][filtered_genes_V2],
                                         gene_info[dataset2]['cell_fracs'][filtered_genes_V2],
                                         log2fc[filtered_genes_V1],
                                         pvals[1]]))
        W_matrices.append(pd.DataFrame([np.repeat(factor, len(top_genes_W)),
                                        top_genes_W,
                                        gene_info[dataset1]['gene_counts'][filtered_genes_V1],
                                        gene_info[dataset2]['gene_counts'][filtered_genes_V1],
                                        gene_info[dataset1]['cell_fracs'][filtered_genes_V1],
                                        gene_info[dataset2]['cell_fracs'][filtered_genes_V1],
                                        log2fc[filtered_genes_V1],
                                        pvals[2]]))
    V1_genes =
    V2_genes =
    W_genes =

    df_cols = ['factor_num', 'gene', 'counts1', 'counts2', 'fracs1', 'fracs2', 'log2fc', 'p_value']
    output_list = [V1_genes, W_genes, V2_genes]

    # cutoff only applies to dataset-specific dfs


    return output_list

"""